from .component import BaseComponent, Node, SimParam, enforce_simparam, SimulationType, _TWO_MAT
from typing import List, Callable
import numpy as np

class NPortS(BaseComponent):
    supported_simulations: List[SimulationType] = [SimulationType.SP]
    
    def __init__(self, nodes: List[Node], gnd: Node, functions: List[List[Callable[[float], float]]], Z0: float):
        super().__init__()
        self.nodes = nodes
        self.nterminals = len(nodes)
        self.gnd = gnd
        self.functions = functions
        self.Z0: SimParam = Z0
        self.component_name = 'NPort(S)'
        self.unit = 'Ohm'
        self.value = self.Z0
    
    @property
    def mat_slice(self) -> tuple[np.ndarray, np.ndarray]:
        ids = np.array([node.index for node in self.nodes] + [self.gnd.index,])
        return np.ix_(ids, ids)

    def SP_matrix_compiler(self):
        row, col = self.mat_slice
        N = self.nterminals

        def comp_function(f: float):
            nF = f.shape[0]
            S = np.array([[sp(f) for sp in row] for row in self.functions], dtype=np.complex128)
            Identity = np.repeat(np.eye(N)[:, :, np.newaxis], nF, axis=2).astype(np.complex128)
            Y = (1/self.Z0) * np.einsum('ijk,jlk->ilk', (Identity-S), np.stack([np.linalg.inv((Identity+S)[:, :, m]) for m in range(nF)],axis=2))
            Y2 = np.zeros((N+1,N+1,nF),dtype=np.complex128)
            Y2[:N,:N,:] = Y
            for i in range(N):
                Y2[i,N,:] = -np.sum(Y[i,:,:],axis=0)
                Y2[N,i,:] = -np.sum(Y[:,i,:],axis=0)
                Y2[N,N,:] += np.sum(Y[i,:,:],axis=0)
            return Y2
        
        def compiler(matrix: np.ndarray, f: float) -> np.ndarray:
            matrix[row,col,:] += comp_function(f)
            return matrix
        
        return compiler
    
class NPortY(BaseComponent):

    supported_simulations: List[SimulationType] = [SimulationType.SP]

    def __init__(self, nodes: List[Node], gnd: Node, functions: List[List[Callable[[float], float]]], Z0: float | SimParam):
        super().__init__()
        self.nodes = nodes
        self.nterminals = len(nodes) + 1
        self.nports = len(nodes)
        self.gnd = gnd
        self.functions = functions
        self.Z0: SimParam = enforce_simparam(Z0)
        self.component_name = 'NPort(Y)'
        self.unit = 'Ohm'
        self.value = self.Y0.value

    @property
    def mat_slice(self) -> tuple[np.ndarray, np.ndarray]:
        ids = np.array([node.index for node in self.nodes] + [self.gnd.index,])
        return np.ix_(ids, ids)

    def SP_matrix_compiler(self):
        row,col = self.mat_slice
        N = self.nterminals
        
        def comp_function(f: float):
            nF = f.shape[0]
            Y = np.array([[y(f) for y in row] for row in self.functions], dtype=np.complex128)
            Y2 = np.zeros((N+1,N+1,nF),dtype=np.complex128)
            Y2[:N,:N,:] = Y
            for i in range(N):
                Y2[i,N,:] = -np.sum(Y[i,:,:],axis=0)
                Y2[N,i,:] = -np.sum(Y[:,i,:],axis=0)
                Y2[N,N,:] += np.sum(Y[i,:,:],axis=0)
            return Y2
        
        def compiler(matrix: np.ndarray, f: float) -> np.ndarray:
            matrix[row,col,:] += comp_function(f)
            return matrix
        
        return compiler
        
        
## NportS
# N = len(nodes)

# def comp_function(f: float):
#     nF = f.shape[0]
#     S = np.array([[sp(f) for sp in row] for row in Sparam], dtype=np.complex128)
#     Identity = np.repeat(np.eye(N)[:, :, np.newaxis], nF, axis=2).astype(np.complex128)
#     Y = (1/Z0) * np.einsum('ijk,jlk->ilk', (Identity-S), np.stack([np.linalg.inv((Identity+S)[:, :, m]) for m in range(nF)],axis=2))
#     Y2 = np.zeros((N+1,N+1,nF),dtype=np.complex128)
#     Y2[:N,:N,:] = Y
#     for i in range(N):
#         Y2[i,N,:] = -np.sum(Y[i,:,:],axis=0)
#         Y2[N,i,:] = -np.sum(Y[:,i,:],axis=0)
#         Y2[N,N,:] += np.sum(Y[i,:,:],axis=0)
#     return Y2
# component = Component(nodes + [gnd, ],[ComponentFunction(comp_function),], Z0).set_metadata(value=Z0, **Components.NPORT)
# self.components.append(component)

## NPORTY




########################################################################################
##
##    DC Solver for the MNA with Numba compilation
##    This file contains the DC solver for the MNA using Numba compilation
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


from numba import njit, prange, c16, i8, f8, typed, types, typeof
import numba_progress as nbp
from numba_progress.progress import ProgressBarType
import numpy as np

#  ___            __  ___    __        __  
# |__  |  | |\ | /  `  |  | /  \ |\ | /__` 
# |    \__/ | \| \__,  |  | \__/ | \| .__/ 
# -------------------------------------------


#placeholder decorator that instead of compiling with numba prints the type using typeof of the called arguments and then executes the function
# def njit(*args, **kwargs):
#     def wrapper(func):
#         def wrapped(*args, **kwargs):
#             print([typeof(arg) for arg in args])
#             return func(*args, **kwargs)
#         return wrapped
#     return wrapper


# Diode data structure
# Tuple[nodeid1, nodeid2, Idata, Vdata]

example = [(1,2,np.array([1.,2.,3.]),np.array([1.,2.,3.])), (4,5,np.array([1.,2.,3.]),np.array([1.,2.,3.]))]



@njit(types.Tuple((f8[:,:],f8[:]))(f8, f8, f8[:], f8[:]), cache=True, parallel=True, fastmath=True)
def compute_diode(v1, v2, Idata, Vdata):
    
    EPS = 1e-15
    dV = v1-v2
    I0 = np.interp(dV, Vdata, Idata)
    i2 = np.interp(dV+EPS, Vdata, Idata)
    i1 = np.interp(dV-EPS, Vdata, Idata)
    G0 = (i2-i1)/(2*EPS)
    G = np.array([[1,-1],[-1,1]]) * G0
    I = np.array([-1,1]) * (I0 - G0*dV)
    return G, I

@njit(f8[:,:](f8[:,:], f8[:,:], i8[:]), cache=True, fastmath=True)
def add_submatrix(larger_matrix, smaller_matrix, indices):
    n = indices.shape[0]
    for i in range(n):
        for j in range(n):
            larger_matrix[indices[i], indices[j]] += smaller_matrix[i, j]
    return larger_matrix

@njit(f8[:](f8[:,:], f8[:], i8, i8, f8[:], i8, typeof(example), f8, f8), 
      cache=True, parallel=True, fastmath=True)
def solve_MNA_DC(Alin, Xlin, N, M, rampsteps, Niter, diodes, eaps, erel):
    """"""

    NM = N+M

    # Initialize the S-parameter matrix
    v_data = np.zeros((N,))
    
    Amat = np.zeros_like(Alin)
    Xvec = np.zeros_like(Xlin)

    diodes = diodes[1:]
    for r in rampsteps:
        for i in range(Niter):
            Amat = Amat*0 + Alin
            Xvec = Xvec*0 + Xlin*r

            for diode in diodes:
                i1, i2 = diode[0], diode[1]
                (G, I) = compute_diode(v_data[i1], v_data[i2], diode[2], diode[3])
                Amat = add_submatrix(Amat, G, np.array([i1,i2]))
                Xvec[i1] = I[0]
                Xvec[i2] = I[1]

            # Reset voltage vector
            v_new = np.zeros((NM,))
            
            v_new[1:] = np.linalg.solve(Amat[1:,1:], Xvec[1:])

            
            # Check convergence
            maxchange = np.max(np.abs(v_new[:N] - v_data[:N]))
            maxV = np.max(np.abs(v_new[:N]))

            if maxchange < eaps + erel*maxV:
                v_data = v_new[:N]
                break
            v_data = v_new[:N]

    return v_data


"""
Sometimes you want to create custom components that have all sorts of internal behavior. There is a class
intended exactly for that!

In this example we will create the effective component of a resistor with parasitics.

"""

import heavi as hv

from heavi.lib.libgen import SubCircuit
import numpy as np

# In case you want to reate an N-port we will use the SubCircuit class.

class MyResistor(SubCircuit):
   
    def __init__(self, resistance: float):
        super().__init__()
        self.n_nodes = 2 # Here we set it manually We must set this property for the class to work.
        self.resistance: float = resistance
       
    """ To make our component do something we have to use the __on_connect__ dunder method"""
    def __on_connect__(self):

        # There are multiple ways to use your own components. In this case we will use a custom impedance function.
        # We will compute the impedance of a resistor with paracitic capacitance and inductance.
        
        # The values of the parasitics are sort of arbitrary for now
        Lps = 10e-9
        Cpp = 0.5e-12
        
        # The function must compute the series impedance
        # as a function of frequency (in hertz)
        def z_parasitic(f):
            # Z_RLC = R + j*2*pi*f*L + 1/(j*2*pi*f*C)
            w = 2 * 3.141592653589793 * f
            Zr = self.resistance
            Zl = 1j * w * Lps
            Zc = 1 / (1j * w * Cpp)
            # return Zr + Zl parallel to Zc
            Z = (Zr+Zl) * Zc / (Zr + Zl + Zc)
            print(Z)
            return Z
        
        # To put our component in the circuit we can do multiple things.
        # The Network object is available as all Node objects carry a reference to it and this method is called after you connected it to nodes.
        # Thus we can use the network to just plug in all components we want. In this case we can just create an impedance component with a function as value.
        self.function = z_parasitic
        
        # our two nodes that this device is connected two are in self.node(1) and self.node(2)
        # Thus if we did
        # MyResistor.connect(n1,n2)
        # Then self.node(1) would return n1 and self.node(2) would return n2 (in order)
        
        # We add the impedance manually with the z_parasitic function for our resistor. We can set the display value to show what the device is.
        # 
        self.network.impedance(self.node(1), self.node(2), z_parasitic, display_value=self.resistance)\
        .set_metadata(name='Custom Component',
                      unit='Ω',
                      value=self.resistance,
                      inductance=Lps,
                      capacitance=Cpp)
        # This metadata can be used to provide more info on what this device is.
        
# Now we can use it!

m = hv.Model()

n1 = m.quick_port(50)
n2 = m.quick_port(50)

# We can create our resistor and connect it to our nodes. We make a simple 50, 50 ohm resistor devider. Its just for the idea of it.
R1 = MyResistor(50)
R1.connect(n1,n2)

R1 = MyResistor(50)
R1.connect(n2,m.gnd)

# The Frequency range
fs = np.logspace(6,10.3,3001)

S = m.run_sp(fs)

hv.plot_s_parameters(fs, [S.S11,S.S21], labels=['S11','S21'], logx=True, dblim=[-20,0])


from .component import BaseComponent, Node, SimParam, enforce_simparam, SimulationType, _TWO_MAT, _TWO_MAT_C
from typing import List, Callable
import numpy as np
from .numeric import stack, cstack

class Port(BaseComponent):
    supported_simulations: List[SimulationType] = [SimulationType.SP,SimulationType.DC,SimulationType.AC]
    SP_sources = 1
    def __init__(self, index: int, sig_node: Node, int_node: Node, gnd_node: Node, Z0: float | SimParam):
        super().__init__()
        self.nodes = [sig_node, int_node, gnd_node]
        self.index: int = index
        self.sig_node: Node = sig_node
        self.int_node: Node = int_node
        self.gnd_node: Node = gnd_node
        self.nterminals = 2
        self.Z0: SimParam = enforce_simparam(Z0)
        self.component_name = 'Port'
        self.unit = 'Ohm'
        self.value = self.Z0.value
    
    @property
    def mat_slice(self) -> tuple[np.ndarray, np.ndarray]:
        ids = np.array([self.sig_node.index, self.int_node.index, self.gnd_node.index, self.sources[0].index])
        return np.ix_(ids, ids)
    
    def SP_matrix_compiler(self):
        row, col = self.mat_slice
        TEMPLATE = np.array([[ 1,-1, 0, 0], 
                             [-1, 1, 0, 0],
                             [ 0, 0, 0, 0],
                             [ 0, 0, 0, 0]])
        TEMPLATE_SOURCE = np.array([[0, 0, 0, 0],
                                    [0, 0, 0, 1],
                                    [0, 0, 0,-1],
                                    [0, 1,-1, 0]])

        def compiler(matrix: np.ndarray, f: float) -> np.ndarray:
            matrix[row,col,:] += stack(TEMPLATE, 1/self.Z0(f), TEMPLATE_SOURCE)
            return matrix
        
        return compiler
    
    def DC_matrix_compiler(self):
        slc = self.mat_slice

        def compiler(matrix: np.ndarray) -> np.ndarray:
            matrix[slc] += 1/self.Z0.value * _TWO_MAT
            return matrix
        
        return compiler
    
    def AC_matrix_compiler(self):
        slc = self.mat_slice

        def compiler(matrix: np.ndarray) -> np.ndarray:
            matrix[slc] += cstack(_TWO_MAT_C, 1/self.Z0.value, 0*_TWO_MAT_C)
            return matrix
        
        return compiler


########################################################################################
##
##    Node Class
##    This file contains the Node class, which is used to represent nodes in the network.
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################


#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------

from __future__ import annotations
from dataclasses import dataclass, field
from typing import Any

#  __             __   __   ___  __  
# /  ` |     /\  /__` /__` |__  /__` 
# \__, |___ /~~\ .__/ .__/ |___ .__/ 
# -------------------------------------------


class UnassignedType:

    def __str__(self):
        return '_'
    
    def __repr__(self):
        return '_'
    
    def __eq__(self, other):
        if other is None:
            return True
        return False

UNASSIGNED = UnassignedType()

@dataclass
class Node:
    """ Node class for the Network object. """
    name: str
    _index: int = field(default=UNASSIGNED)
    _parent: Any = None
    _linked: Node = None
    _gnd: bool = False

    def __repr__(self) -> str:
        if self._gnd:
            return 'Node_GND'
        if self._linked is None:
            return f"{self.name}[{self._index}]"
        else:
            return f"LinkedNode[{self._linked}]"
    
    def __str__(self) -> str:
        return self.__repr__()
    
    def __hash__(self):
        return hash(f'{self.name}_{self.index}')
    
    def set_index(self, index: int):
        """Assigns an index to the node. Should only be used by the Network object.

        Args:
            index (int): The node index.
        """        
        self._index = index

    def unique(self) -> Node:
        """Unique returns the unique variant of the node. If the node is linked to another node, 
        the linked node is returned. If the node is not linked, the node itself is returned.

        Returns:
            Node: The unique node.
        """        
        if self._linked is not None:
            return self._linked
        return self
    
    @property
    def index(self) -> int:
        """Returns the index of the node. If the node is linked to another node, 
        the index of the linked node is returned.

        Returns:
            int: the node index
        """        
        if self._linked is not None:
            return self._linked.index
        return self._index

    def merge(self, other: Node) -> Node:
        """Merges the current node with another node. The node that is merged with is returned."""
        self._linked = other
        return self
    
    def __gt__(self, other: Node) -> Node:
        """Merges the current node with another node. The node that is merged with is returned."""
        if isinstance(other, Node):
            self._linked = other
            return other
        return NotImplemented


@dataclass
class MatSource:
    """The MatSource class is used to represent a source in the network.
    This class is used only to store indices of required extra source objects during
    Matrix construction. Each MNA matrix requires additional collumns and rows for each
    Source object. Components that require additional sources to represent their behavior
    should return a MatSource object with the required indices during the matrix construction
    """    
    index: int

class NodeAllocator:

    def __init__(self, index: str, parent):
        self._index: str = index
        self._pointer: int = 0
        self._nodes: dict[int, Node] = dict()
        self._parent = parent
    
    def _get_new_number(self) -> int:
        """Returns the first available number for a new node.

        Returns:
            int: The new node number
        """        
        number = self._pointer
        while True:
            self._pointer += 1
            if self._pointer not in self._nodes:
                break
        return number
    
    def _make(self, number: int) -> Node:
        """Creates a node with the specified number.

        Args:
            number (int): the index of the node

        Returns:
            Node: The new node object
        """        
        node_obj = Node(f'{self._index}{number}',_parent = self._parent)
        self._nodes[number] = node_obj
        return node_obj
                    
    def request(self, number: int = None) -> Node:
        """Requests a node with the specified number. If the node does not exist, it is created.
        If no number is specified, a new number is generated. If the node exists, it is returned.

        Args:
            number (int, optional): The node index. Defaults to None.
        
        Returns:
            Node: The requested node object.
        """
        new = False
        if number is None:
            number = self._get_new_number()
        if number not in self._nodes:
            self._make(number)
            new = True
        return self._nodes[number], new


"""This script will demonstarte the basics of Heavi.

We will be making a simple 
"""

# First we import heavi

import heavi as hv

# Next we create our circuit. We can use the Network class
# but its better to use the more extended Model class

m = hv.Model()

# Heavi works by creating nodes and connecting components to them.
# The basic library of lumped elements allows us to directly connect them
# in the constructor of the components.
# More advanced components have extra methods. More about that later.

# We always have our ground node
gnd = m.gnd

# For this filter we will make two ports. To create a terminal
# for S-parameter analysis we need intermediate nodes and such thus we make this
# easier by using the build in methods!

node_in = m.node()
p1 = m.new_port(50, node_in)

# If we are lazy, we can also make the port and the node at the same time. 
node_out = m.quick_port(50)

# Our terminal object is now not available as before but we don't use those anyway.

# Next for our filter we need two intermediate nodes:
n1 = m.node()
n2 = m.node()

# Now we can connect some components
L1 = m.inductor(node_in, n1, 2.407e-9)
C1 = m.capacitor(n1, gnd, 1.66145e-12)
L2 = m.inductor(n1, n2, 5.0207e-9)
C2 = m.capacitor(n2, gnd, 1.55145e-12)
L3 = m.inductor(n2, node_out, 2.407e-9)

# Now that our modeling is done we can simply run our circuit!
fs = hv.frange(0.5e9, 8e9, 2001)
S = m.run_sp(fs)

# Our S-matrix object is an Sparameter object.

hv.plot_s_parameters(fs, [S.S11, S.S21, S.S12, S.S22], labels=['S11','S21','S12','S22'], logx=True)



#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


from __future__ import annotations
from .libgen import SubCircuit
import numpy as np
from scipy.interpolate import interp1d


class EMergeModel(SubCircuit):
    """An EMerge N-port model

    """
    def __init__(self, emerge_data: tuple[np.ndarray, np.ndarray]
                 ,interp_type: str = 'cubic'):
        super().__init__()
        
        self.Fdata, self.Sdata = emerge_data
        
        self.n_ports: int = self.Sdata.shape[1]
        self.n_nodes = self.n_ports
        self.interp_type: str = interp_type
        self.refimp: float = 50
        
    def renormalize(self, new_impedance: float) -> EMergeModel:
        """ Renormalize the touchstone file to a new impedance.

        Args:
            new_impedance (float): The new impedance to normalize the touchstone file to.

        Returns:
            FileBasedNPort: The renormalized touchstone file subcircuit.
        """        
        Zold = self.refimp
        Znew = new_impedance
        A = np.eye(self.n_ports, self.n_ports) * np.sqrt(Znew/Zold)*(1/(Znew+Zold))
        R = np.eye(self.n_ports, self.n_ports) * ((Znew-Zold)/(Znew+Zold))
        iA = np.linalg.pinv(A)
        E = np.eye(self.n_ports, self.n_ports)
        for iif in range(self.n_frequencies):
            S = self.Sdata[iif,:,:]
            self.Sdata[iif,:,:] = iA @ (S - R) @ np.linalg.pinv(E - R @ S) @ A
        self.refimp = new_impedance
        return self

    def __on_connect__(self):
        s_functions = []
        for i in range(self.n_ports):
            row = []
            for j in range(self.n_ports):
                row.append(interp1d(self.Fdata, self.Sdata[:,i,j], kind=self.interp_type, bounds_error=False, fill_value=(self.Sdata[0,i,j],self.Sdata[-1,i,j])))
            s_functions.append(row)
        

        self.network.n_port_S(self.network.gnd, 
                              [self.nodes[i] for i in range(1,self.n_ports+1)],s_functions,self.refimp)
        



""" Monte Carlo Simulations

In this example we will show you how to do Monte-Carlo simulations.
"""
import heavi as hv
import numpy as np

M = hv.Model()

# Just like with Parameter sweeps we create a MonteCarlo object which will manage our parameter variations.

mc = hv.MonteCarlo()

# We will create a Capacitor value that is distributed by a Gaussian function.

Cvalue = mc.gaussian(1e-12, 1e-13) # Gaussian distribution with mean=1pF, standard deviation=0.1pF

# Next we create our network with a ristor and capacitor
n1 = M.quick_port(50)
n2 = M.quick_port(50)

M.capacitor(n1,n2,Cvalue)
M.resistor(M.gnd,n2,10_000)

# And we simulate our model
f = hv.frange(1e6,3e9,1001)

# We use the iterate funtion to run our model for 1000 steps
Ss = []
for itt in mc.iterate(201):
    S = M.run_sp(f)
    Ss.append(S)

# And we do some post processing to plot the range of values we may expect.
Ss = np.array([s.S21 for s in Ss])

Smin = np.min(Ss, axis=0)
Smax = np.max(Ss, axis=0)
Smean = np.mean(Ss, axis=0)

hv.plot_s_parameters(f, [Smin, Smean, Smax], labels=['S21 min','S21','S21 max'], linestyles=['--','-','--'], colorcycle=[0,0,0])


########################################################################################
##
##    Default Source Components
##    These components are used to define the behavior of voltage sources in the circuit
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


from .component import BaseComponent, Node, SimParam, enforce_simparam, SimulationType, _TWO_MAT
from typing import List, Callable
import numpy as np


#  __             __   __   ___  __  
# /  ` |     /\  /__` /__` |__  /__` 
# \__, |___ /~~\ .__/ .__/ |___ .__/ 
# -------------------------------------------


class VoltageDC(BaseComponent):
    supported_simulations: List[SimulationType] = [SimulationType.AC,
                                                   SimulationType.DC,
                                                   SimulationType.SP,]
    
    DC_sources = 1
    
    def __init__(self, node1: Node, node2: Node, voltage: float, r_series: float | SimParam = 0.001):
        """Creates a DC voltage source component

        Args:
            node1 (Node): The positive terminal of the voltage source
            node2 (Node): The negative terminal of the voltage source
            voltage (float): The voltage of the source
            r_series (float | SimParam, optional): The series impedance. Defaults to 0.001.
        """        
        super().__init__()
        self.nodes = [node1, node2]
        self.nterminals = 2
        self.V: float = voltage
        self.r_series: SimParam = enforce_simparam(r_series)
        self.component_name = 'DC Voltage Source'
        self.unit = 'V'
        self.value = self.V
    
    @property
    def mat_slice(self) -> tuple[np.ndarray, np.ndarray]:
        ids = np.array(self.indices())
        return np.ix_(ids, ids)
    
    @property
    def vec_slice(self) -> np.ndarray:
        return np.array(self.indices())
    
    @property
    def index(self) -> int:
        return self.sources[0].index
    
    def DC_matrix_compiler(self):
        
        MAT_TEMPLATE = np.array([[0, 0, -1],
                             [0, 0, 1],
                             [-1, 1, 0]])
        slc = self.mat_slice

        def compiler(matrix: np.ndarray):
            matrix[slc] += MAT_TEMPLATE
            return matrix
        
        return compiler
    
    def DC_source_compiler(self):
        voltage = self.V
        index = self.sources[0].index

        def compiler(vector: np.ndarray):
            vector[index] = voltage
            return vector
        return compiler
    
    def SP_matrix_compiler(self):
        slc = self.mat_slice

        def compiler(matrix: np.ndarray, f: float) -> np.ndarray:
            matrix[slc] += self.r_series(f) * _TWO_MAT
            return matrix
        
        return compiler
    
    def AC_matrix_compiler(self):
        return self.SP_matrix_compiler()
    

class VoltageFN(BaseComponent):
    supported_simulations: List[SimulationType] = [SimulationType.TRANS,]
    
    TRANS_sources = 1
    
    def __init__(self, node1: Node, node2: Node, voltage_function: float, r_series: float | SimParam = 0.001):
        """Creates a transient voltage source component"

        Args:
            node1 (Node): The positive terminal of the voltage source
            node2 (Node): The negative terminal of the voltage source
            voltage_function (float): The voltage function of the source
            r_series (float | SimParam, optional): The series impedance. Defaults to 0.001.
        
        """
        super().__init__()
        self.nodes = [node1, node2]
        self.nterminals = 2
        self.vfun: Callable = voltage_function
        self.r_series: SimParam = enforce_simparam(r_series)
        self.component_name = 'Transient Voltage Source'
        self.unit = 'V'
        self.value = 0
    
    @property
    def index(self) -> int:
        return self.sources[0].index
    
    @property
    def mat_slice(self) -> tuple[np.ndarray, np.ndarray]:
        ids = np.array(self.indices())
        return np.ix_(ids, ids)
    
    @property
    def vec_slice(self) -> np.ndarray:
        return np.array(self.indices())
    
    def generate_timeseries(self, t: np.ndarray) -> np.ndarray:
        return self.vfun(t)
    
    def TRANS_matrix_compiler(self):
        
        MAT_TEMPLATE = np.array([[0, 0, -1],
                             [0, 0, 1],
                             [-1, 1, 0]])
        slc = self.mat_slice

        def compiler(matrix: np.ndarray):
            matrix[slc] += MAT_TEMPLATE
            return matrix
        
        return compiler
    
    def SP_matrix_compiler(self):
        slc = self.mat_slice
        # EFFECTIVE SHORT CIRCUIT
        def compiler(matrix: np.ndarray, f: float) -> np.ndarray:
            matrix[slc] += self.r_series(f) * _TWO_MAT
            return matrix
        
        return compiler
    
    def DC_matrix_compiler(self):
        return self.SP_matrix_compiler()
    
    def AC_matrix_compiler(self):
        return self.SP_matrix_compiler()


########################################################################################
##
##    Spice-Component import modules
##    This module is used to import SPICE subcircuits and generate Python classes for them.
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################


#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


from __future__ import annotations

from pathlib import Path
from .libgen import TwoNodeSubCircuit
from ..network import Node
import re
from dataclasses import dataclass
from typing import List, Optional

from loguru import logger

#  __   __        __  ___           ___  __  
# /  ` /  \ |\ | /__`  |   /\  |\ |  |  /__` 
# \__, \__/ | \| .__/  |  /~~\ | \|  |  .__/ 
# -------------------------------------------


_POWERS = {
        "f": -15,
        "p": -12,
        "n": -9,
        "u": -6,
        "m": -3,
        "k": 3,
        "q": 6,
        "g": 9,
        "t": 12,
    }

_COMPONENT_TEMPLATE: str = '''


class #NAME#(BaseTwoPort):
    """ #DESCRIPTION# """
    def __init__(self):
        super().__init__()
    
    def __on_connect__(self):
        node1 = self.node(1)
        node2 = self.node(2)
        
#NODES#
        
#REST#'''

_PYTHON_BASE_TEMPLATE = '''
from __future__ import annotations
from heavi.lib.libgen import BaseTwoPort
from heavi.rfcircuit import Node


'''

#  __             __   __   ___  __  
# /  ` |     /\  /__` /__` |__  /__` 
# \__, |___ /~~\ .__/ .__/ |___ .__/ 
# -------------------------------------------


class SpaceCompContainer:
    _implemented: bool = False
    def __init__(self, name: str, nodes: List[int], value: float):
        self.name = name
        self.nodes = nodes
        self.value = value

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name}, nodes={self.nodes}, value={self.value})"
    
class SpiceR(SpaceCompContainer):
    _implemented: bool = True

class SpiceL(SpaceCompContainer):
    _implemented: bool = True

class SpiceC(SpaceCompContainer):
    _implemented: bool = True

class SpiceV(SpaceCompContainer):
    _implemented: bool = False

class SpiceI(SpaceCompContainer):
    _implemented: bool = False

class SpiceX(SpaceCompContainer):
    _implemented: bool = False

class SpiceD(SpaceCompContainer):
    _implemented: bool = False

class SpiceSubCircuit:

    def __init__(self, name: str, nodes: List[int], components: List[SpaceCompContainer]):
        self.name = name
        self.nodes = nodes
        self.components = components
        
    def __repr__(self):
        return f"Subcircuit(name={self.name}, nodes={self.nodes}, components={self.components})"
    
    @property
    def is_supported(self) -> bool:
        ''' Returns True if all components in the subcircuit are supported by the compiler. '''
        for component in self.components:
            if not component._implemented:
                return False
        return True
    
    def all_containing_nodes(self) -> list[int]:
        ''' Returns a list of all nodes mentioned in the subcircuit. '''
        nodes = self.nodes
        for component in self.components:
            nodes.extend(component.nodes)
        return sorted(list(set(nodes)))


class SpiceASY:

    def __init__(self, description: list[str], spicemodel: str, instname: str, model_file: str):
        self.description = description
        self.spicemodel = spicemodel
        self.instname = instname
        self.model_file = model_file

        if self.instname is None:
            self.instname = 'unnamed'
        if self.model_file is None:
            self.model_file = 'unnamed'
    

def _parse_asy_file(file_path: str) -> SpiceASY:
    ''' Parses a .asy file and returns a list of all components. 
    
    parses the following datalines:
    SYMATTR Description WE-MPSA EMI Multilayer Power Suppression Bead \nAfter inserting, right-click on the symbol to select the part number. \nwww.we-online.com/catalog/WE-MPSA \n\nPlease note disclaimer in WE-MPSA.lib.
    SYMATTR SpiceModel 0805_78279220321_320Ohm
    SYMATTR Prefix x
    SYMATTR InstName L
    SYMATTR ModelFile WE-MPSA.lib
    '''

    
    components = []
    description = None
    spicemodel = None
    instname = None
    model_file = None

    with open(file_path, 'r', encoding='utf-8', errors='replace') as f:
        lines = f.readlines()
        for line in lines:
            if line.startswith("SYMATTR Description"):
                description = line.split("SYMATTR Description ")[1].strip().replace("\\n", "\n")
            elif line.startswith("SYMATTR SpiceModel"):
                spicemodel = line.split("SYMATTR SpiceModel ")[1].strip()
            elif line.startswith("SYMATTR InstName"):
                instname = line.split("SYMATTR InstName ")[1].strip()
            elif line.startswith("SYMATTR ModelFile"):
                model_file = line.split("SYMATTR ModelFile ")[1].strip()
    
    return SpiceASY(description, spicemodel, instname, model_file)
              

def merge_continuation_lines(file_path: str) -> List[str]:
    """
    Reads the file and merges lines that begin with '+' with the previous line.
    """
    merged_lines = []
    with open(file_path, 'r', encoding='utf-8', errors='replace') as f:
        current_line = ""
        for line in f:
            # Remove trailing newline and spaces.
            line = line.rstrip()
            # If the line starts with '+' it is a continuation of the previous line.
            if line.startswith("+"):
                current_line += " " + line.lstrip("+").strip()
            else:
                # If we have a previous line, add it to the list.
                if current_line:
                    merged_lines.append(current_line)
                current_line = line.strip()
        if current_line:
            merged_lines.append(current_line)
    return merged_lines


def parse_token(string: str, numdict: dict) -> float:
    ''' Parses a string with a number and a power of 10. '''
    if string[0] == "{" and string[-1] == "}":
        return numdict.get(string[1:-1], None)
    else:
        string = re.sub(r'(?i)\bmeg\b', 'M', string)
        groups = re.match(r'([-+]?\d*\.?\d+(?:[eE][-+]?\d+)?)([a-zA-Z]?)([a-zA-Z]*)?', string)
        # Groups is Number, potential power and unit. Units are ignored
        if groups is None:
            raise ValueError(f"Invalid number: {string}, {numdict}")
        number = float(groups.group(1))
        power = groups.group(2)
        if power:
            power = _POWERS.get(power.lower(), 0)
            number *= 10 ** power
        return number

def parse_subckt_line(line: str) -> SpiceSubCircuit:
    # Remove excess spaces and normalize whitespace
    line = re.sub(r"\s+", " ", line.strip()).upper()

    # Match the .subckt line structure
    match = re.match(r"\.SUBCKT\s+([a-zA-Z\d\-\_\.\+]+)\s+([0-9a-zA-Z\s]+)(?:\s+PARAMS:\s*(.*))?", line)
    if not match:
        raise ValueError(f"Invalid .subckt line: {line}")
    name = match.group(1)  # Subcircuit name
    nodes = list(match.group(2).split()) # Convert nodes to integers
    params_str = match.group(3) if match.group(3) else ""  # Parameter string

    # Extract parameters using regex (matches key=value pairs)
    params = {}
    param_matches = re.findall(r"(\w+)=([\d.eE+-]+[a-zA-Z]?)", params_str)

    

    for key, value in param_matches:
        params[key] = parse_token(value,{}) # Direct conversion

    return name, nodes, params

def parse_spice_lib(file_path: str) -> List[SpiceSubCircuit]:
    """
    Parses a SPICE .lib file (with or without continuation lines)
    and returns a list of Subcircuit objects ONLY if the subcircuit contains
    components whose names begin with R, L, C, V, or I.
    
    Each Subcircuit object contains:
      - name: the subcircuit name (from the .subckt header)
      - nodes: the list of nodes declared in the header
      - components: a list of component objects (SpiceR, SpiceL, SpiceC, etc.)
    
    If any component in the subcircuit does not begin with one of the allowed
    letters, that entire subcircuit is skipped.
    """
    allowed_types = {"R", "L", "C", "V", "I"}
    merged_lines = merge_continuation_lines(file_path)
    subcircuits: List[SpiceSubCircuit] = []
    current_subckt: Optional[str] = None
    current_nodes: List[str] = []
    current_components: List[SpiceComponent] = []
    valid_subckt = True  # Flag to indicate that current subckt contains only allowed components
    
    params = {}
    for iline, line in enumerate(merged_lines, start=1):
        # Skip comment lines.
        if line.startswith("*") or not line:
            continue
        
        lower_line = line.lower()
        if lower_line.startswith(".subckt"):
            # If we were in a subckt (without seeing .ends) we could handle it here.
            current_subckt, current_nodes, params = parse_subckt_line(line)
            # tokens[0] is ".subckt", tokens[1] is the subckt name, tokens[2:] are node names.
            valid_subckt = True  # start with a valid subcircuit assumption
        elif lower_line.startswith(".ends"):
            if current_subckt is not None and valid_subckt:
                subcircuit_obj = SpiceSubCircuit(name=current_subckt,
                                             nodes=current_nodes,
                                             components=current_components)
                subcircuits.append(subcircuit_obj)
            # Reset for the next subcircuit.
            current_subckt = None
            current_nodes = []
            current_components = []
            valid_subckt = True
        else:
            # We are inside a subcircuit definition.
            if current_subckt is not None:

                #removed_statements = re.sub(r"\{.*?\}", "", line)
                removed_statements = re.sub(r"\{\w+=\w+\}", "", line)
                tokens = [token for token in removed_statements.split() if token]
                # Expect at least 4 tokens: component name, node1, node2, value.
                if len(tokens) < 4:
                    logger.error(f"Error parsing component: {line} on line number {iline} in file {file_path}. Insufficient information detected")
                    continue  # Or optionally raise an error.
                
                comp_name = tokens[0]
                comp_type = comp_name[0].upper()
                
                node1 = tokens[1]
                node2 = tokens[2]

                if comp_type not in allowed_types:
                    valid_subckt = False
                    break
                
                try:
                    value = parse_token(tokens[3], params)
                except ValueError:
                    logger.error(f"Error parsing component: {line} on line number {iline} in file {file_path}")
                    valid_subckt = False
                    break
                
                # Create the appropriate component instance.
                try:
                    if comp_type == "R":
                        comp_obj = SpiceR(name=comp_name, nodes=[node1, node2], value=value)
                    elif comp_type == "L":
                        comp_obj = SpiceL(name=comp_name, nodes=[node1, node2], value=value)
                    elif comp_type == "C":
                        comp_obj = SpiceC(name=comp_name, nodes=[node1, node2], value=value)
                    elif comp_type == "V":
                        comp_obj = SpiceV(name=comp_name, nodes=[node1, node2], value=value)
                    elif comp_type == "I":
                        comp_obj = SpiceI(name=comp_name, nodes=[node1, node2], value=value)
                    elif comp_type == "X":
                        comp_obj = SpiceX(name=comp_name, nodes=[node1, node2], value=value)
                    elif comp_type == "D":
                        comp_obj = SpiceD(name=comp_name, nodes=[node1, node2], value=value)
                    else:
                        # Should not get here because of the allowed_types check.
                        valid_subckt = False
                        continue
                except ValueError:
                    logger.error(f"Error parsing component: {line} on line number {iline} in file {file_path}")
                    valid_subckt = False
                    continue
                current_components.append(comp_obj)
    return subcircuits

class SpiceComponent(TwoNodeSubCircuit):

    def __init__(self, subcircuit: SpiceSubCircuit):
        super().__init__()
        self.subcircuit = subcircuit

    def add_component(self, component: SpaceCompContainer, node1: Node, node2: Node):
        value = component.value
        if isinstance(component, SpiceR):
            self.network.resistor(node1, node2, value)
        elif isinstance(component, SpiceL):
            self.network.inductor(node1, node2, value)
        elif isinstance(component, SpiceC):
            self.network.capacitor(node1, node2, value)
        elif isinstance(component, SpiceV):
            self.network.voltage_source(node1, node2, value)
        elif isinstance(component, SpiceI):
            self.network.current_source(node1, node2, value)

    def __on_connect__(self):
        node1 = self.node(1)
        node2 = self.node(2)

        mentioned_nodes = self.subcircuit.all_containing_nodes()

        node_map = {self.subcircuit.nodes[0]: node1, self.subcircuit.nodes[1]: node2}
        for i in mentioned_nodes:
            if i not in node_map:
                node_map[i] = self.network.node(f'{self.subcircuit.name}[{i}]')
        
        for component in self.subcircuit.components:
            node1 = node_map[component.nodes[0]]
            node2 = node_map[component.nodes[1]]
            print(f"Adding component {component.name} with value {component.value} between nodes {node1} and {node2}")
            self.add_component(component, node1, node2)


class SpiceLibrary:

    def __init__(self, filepath: str):
        self.filepath = filepath
        self.subcircuits: list[SpiceSubCircuit] = parse_spice_lib(filepath)
        self.circuit_dict: dict[str, SpiceSubCircuit] = {subckt.name.lower(): subckt for subckt in self.subcircuits}
    
    def merge(self, other: SpiceLibrary) -> SpiceLibrary:
        ''' Merge another SpiceLibrary objects into the current one. '''
        self.subcircuits.extend(other.subcircuits)
        self.circuit_dict.update(other.circuit_dict)
        return self
    
    def __call__(self, searchterm: str) -> Optional[SpiceComponent]:
        sbk = self.find(searchterm)
        if sbk is None:
            return None
        return SpiceComponent(sbk)
    
    def find(self, name: str) -> Optional[SpiceSubCircuit]:
        name = name.lower()

        if name in self.circuit_dict:
            return self.circuit_dict[name]
        else:
            logger.warning(f"No subcircuits found for '{name}'")
            return None


def import_spice_library_directory(directory: str, libfilename: str = None) -> None:
    """Imports a directory with .lib and .asy files and generates a Python file with the components.
    This function searches all .lib files in the provided directory to build one big SpiceLibrary object.
    Then it searches foor all spice .asy files to generate a set of components. 
    Based on this it generates a Python file with the components and the SpiceLibrary object.

    Args:
        directory (str): The directory to search for .lib and .asy files.
        libfilename (str, optional): The output directory name. Defaults to None.
    """    

    lib_files = Path(directory).rglob("*.lib")
    asy_files = Path(directory).rglob("*.asy")

    spicelib = SpiceLibrary(next(lib_files))

    for file in lib_files:
        spicelib.merge(SpiceLibrary(file))
    
    asys = [_parse_asy_file(file) for file in asy_files]

    pythonlines = _PYTHON_BASE_TEMPLATE

    for asy in asys:
        sbk = spicelib.find(asy.spicemodel)

        if sbk is None:
            # Log warning already thrown in the .find function
            continue

        if not sbk.is_supported:
            logger.warning(f"Subcircuit '{sbk.name}' contains unsupported components. Skipping this in outputting.")
            continue
        
        sbknode1 = sbk.nodes[0]
        sbknode2 = sbk.nodes[1]
        
        # Generate the component class name by removing illegal python characters.
        component_class_name = asy.instname.upper() + sbk.name.replace(" ", "").replace('.','p').replace('-','_')

        #Substitute the class name
        component_text = _COMPONENT_TEMPLATE.replace("#NAME#", component_class_name)
        
        #Substitute the description
        component_text = component_text.replace("#DESCRIPTION#", f"Component for {asy.spicemodel}\n{asy.description}")
        
        mentioned_nodes = sbk.all_containing_nodes()
        
        node_text = []
        # Define node statements for each node in the subcircuit that isn't also the main subcircuit connectio nodes.
        for node in mentioned_nodes:
            if node == sbknode1 or node == sbknode2:
                continue
            node_text.append(f"        N_{node} = self.network.node('{sbk.name}[{node}]')")
        component_text = component_text.replace("#NODES#", "\n".join(node_text))

        # Define the component statements for each component in the subcircuit.
        component_lines = []
        for component in sbk.components:
            # Define the cubcircuit component nodes
            node1 = component.nodes[0]
            node2 = component.nodes[1]

            nn1 = f"N_{node1}"
            nn2 = f"N_{node2}"
            
            # Replace the nodes by the two primary nodes of the subcircuit
            if node1 == sbknode1:
                nn1 = "node1"
            elif node1 == sbknode2:
                nn1 = "node2"
            if node2 == sbknode1:
                nn2 = "node1"
            elif node2 == sbknode2:
                nn2 = "node2"

            fname = ''
            if isinstance(component, SpiceR):
                fname = 'resistor'
            elif isinstance(component, SpiceL):
                fname = 'inductor'
            elif isinstance(component, SpiceC):
                fname = 'capacitor'

            component_lines.append(f"        self.network.{fname}({nn1}, {nn2}, {component.value})")
        
        component_text = component_text.replace("#REST#", "\n".join(component_lines))
        pythonlines += component_text

        with open(f"{libfilename}.py", "w") as f:
            f.write(pythonlines)


########################################################################################
##
##    Numerical Features
##    This module contains classes used to represent numerical values that may change
##    value in various conditions.
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


from __future__ import annotations
import numpy as np
from typing import Callable, Generator
from itertools import product
from .sparam import NDSparameters, StatSparameters, StatSparam
import inspect
from functools import wraps

#  __   __        __  ___           ___  __  
# /  ` /  \ |\ | /__`  |   /\  |\ |  |  /__` 
# \__, \__/ | \| .__/  |  /~~\ | \|  |  .__/ 
# -------------------------------------------


TEN_POWERS = {
    -12: "p",
    -9: "n",
    -6: "u",
    -3: "m",
    0: "",
    3: "k",
    6: "M",
    9: "T",
    12: "P",
}

#  __   __             ___         ___       __   ___     ___            __  ___    __        __  
# /  ` /  \ |\ | \  / |__  |\ | | |__  |\ | /  ` |__     |__  |  | |\ | /  `  |  | /  \ |\ | /__` 
# \__, \__/ | \|  \/  |___ | \| | |___ | \| \__, |___    |    \__/ | \| \__,  |  | \__/ | \| .__/ 
# -------------------------------------------


def _get_power(number: float) -> tuple[float, int]:
    """Returns the base and exponent in scientific notation.

    Args:
        number (float): The number to convert.

    Returns:
        tuple[float, int]: The base and exponent.
    """    
    ten_power = np.log10(np.abs(number))
    n_thousands = np.floor(ten_power / 3) * 3
    n_thousands = min(12, max(-12, n_thousands))
    base_number = number / (10**n_thousands)
    return base_number, n_thousands

def format_value_units(value: float, unit: str) -> str:
    """Formats a value with units in scientific notation.

    Args:
        value (float): The value to format.
        unit (str): The unit of the value.

    Returns:
        str: The formatted value.

    Example:
    >>> format_value_units(1e-9, "A")
    "1.00 nA"
    """    
    
    v, p = _get_power(value)
    return f"{v:.2f} {TEN_POWERS[p]}{unit}"

class Uninitialized:
    """A class to represent an uninitialized value"""
    def __init__(self):
        pass

    def __repr__(self):
        return "Uninitialized"
    
    def __call__(self, f):
        raise ValueError("Uninitialized value")
    
    def __mul__(self, other):
        raise ValueError("Uninitialized value")
    
    def __add__(self, other):
        raise ValueError("Uninitialized value")
    
    def __sub__(self, other):
        raise ValueError("Uninitialized value") 
    

class SimParam:
    """A class to represent a simulation parameter.

    """    
    _eval_f = 1e9

    def __init__(self, unit=''):
        self.unit = ''
        self._inf = np.inf
        

    @property
    def value(self) -> float:
        return self(0)
    
    def scalar(self) -> float:
        """Returns the scalar value of the parameter"""
        return self.value
    
    def initialize(self) -> None:
        """Initializes the parameter"""
        pass

    def __call__(self, f: np.ndarray) -> np.ndarray:
        """Returns the value of the parameter at a given frequency"""
        return np.nan_to_num(np.ones_like(f) * self._value, posinf=self._inf, neginf=-self._inf)
    
    def __repr__(self) -> str:
        """Returns a string representation of the parameter"""
        return f"SimParam({self.value})"
    
    def negative(self) -> SimParam:
        """Returns the negative of the parameter"""
        return Function(lambda f: -self.sp(f))
    
    def inverse(self) -> SimParam:
        """Returns the inverse of the parameter"""
        return Function(lambda f: 1/self.sp(f))
    
class Scalar(SimParam):
    """A class to represent a scalar value"""

    def __init__(self, value: float, unit: str = '', inf: float = np.inf):
        self._value = value
        self._inf = inf
        self.unit = unit
    
    def __repr__(self) -> str:
        return f"SimValue({self._value})"
    
    def negative(self) -> SimParam:
        """Returns the negative of the scalar"""
        return Scalar(-self._value)
    
    def inverse(self) -> SimParam:
        """Returns the inverse of the scalar"""
        return Scalar(1/self._value)

    
class Negative(SimParam):
    """A class to represent the negative of a parameter"""

    def __init__(self, simparam: SimParam, inf: float = np.inf):
        self._simparam: SimParam = simparam
        self.unit = simparam.unit
        self._inf = inf

    def __repr__(self) -> str:
        return f"Negative({self._simparam})"
    
    def __call__(self, f: np.ndarray) -> np.ndarray:
        return np.nan_to_num(-self._simparam(f), posinf=self._inf, neginf=-self._inf)
    
    def negative(self) -> SimParam:
        """Returns the negative of the negative"""
        return self._simparam

class Inverse(SimParam):
    
    def __init__(self, simparam: SimParam, inf: float = np.inf):
        self._simparam: SimParam = simparam
        self.unit = simparam.unit
        self._inf = inf

    def __repr__(self) -> str:
        return f"Inverse({self._simparam})"
    
    def __call__(self, f: np.ndarray) -> np.ndarray:
        return np.nan_to_num(1/self._simparam(f), posinf=self._inf, neginf=-self._inf)
    
    def inverse(self) -> SimParam:
        """Returns the inverse of the inverse"""
        return self._simparam
    
class Function(SimParam):

    def __init__(self, 
                 function: Callable[[np.ndarray], np.ndarray], 
                 unit: str ='',
                 inf: float = np.inf):
        """A class to represent a function of frequency.
        
        Parameters:
        -----------
        function : Callable[[np.ndarray], np.ndarray]
            The function of frequency.
        """
        self._func = function
        self._inf = inf
        self.unit: str = unit

    def __call__(self, f: np.ndarray) -> np.ndarray:
        return np.nan_to_num(self._func(f), posinf=self._inf, neginf=-self._inf)

    def __repr__(self) -> str:
        return f"Function({self._func})"
    
    
class Random(SimParam):

    def __init__(self, randomizer: Callable, unit: str = '', inf: float = np.inf):
        """A class to represent a random value.
        
        Parameters:
        -----------
        randomizer : Callable
            A function that returns a random value.
        """
        super().__init__()
        self._randomizer = randomizer
        self._value = Uninitialized()
        self._mean = None
        self._std = None
        self._inf = inf
        self.unit = unit

    @property
    def value(self) -> float:
        return self._value
    
    def initialize(self):
        self._value = self._randomizer()

    def __repr__(self):
        return f"Gaussian({self._mean}, {self._std})"   
    
    def negative(self) -> SimParam:
        return Negative(self)
    
    def inverse(self) -> SimParam:
        return Inverse(self)

            
class Param(SimParam):
    """ A class to represent a parameter that is swept over a range of values.

    Parameters:
    -----------
    values : np.ndarray
        An array of values to sweep over
    """
    def __init__(self, values: np.ndarray, unit: str = '', inf: float = np.inf):
        super().__init__()
        self._values = values
        self._index: int = 0
        self._value = Uninitialized()
        self._inf = inf
        self.unit: str = unit

    @staticmethod
    def lin(start: float, stop: float, Nsteps: int) -> Param:
        """Creates a linearly spaced parameter sweep."""
        return Param(np.linspace(start,stop,Nsteps))
    
    @staticmethod
    def range(start: float, stop: float, step: float, *args, **kwargs) -> Param:
        """Creates a range of values with a given step size."""
        return Param(np.arange(start, stop, step, *args, **kwargs))

    def __len__(self):
        return len(self._values)
    
    def __repr__(self):
        ## shortened list of values (start and end only)
        return f"Param({self._values[0]}, ..., {self._values[-1]})"
    
    def __call__(self, f):
        return np.nan_to_num(self._value * np.ones_like(f), posinf=self._inf, neginf=-self._inf)
    
    def set_index(self, index: int):
        """Sets the index of the parameter.
        
        Parameters:
        -----------
        index : int
            The index of the parameter.
        """

        self._index = index
    
    def initialize(self):
        self._value = self._values[self._index]
    
    def negative(self) -> SimParam:
        return Negative(self)
    
    def inverse(self) -> SimParam:
        return Inverse(self)


class ParameterSweep:
    """A class to represent a parameter sweep."""
    def __init__(self):
        self.sweep_dimensions: list[tuple[Param]] = []
        self.index_series: list[tuple[int]] = []
        self._index: int = 0
        self._param_buffer: list = []
        self._S_data: list[np.ndarray] = []
        self._current_index: tuple[int] = None
        self._mdim_data: NDSparameters = None
        self._fdata: np.ndarray = None
    
    def lin(self, start: float, stop: float) -> ParameterSweep:
        """Adds a linear sweep of values to the parameter sweep.
        
        Parameters:
        -----------
        start : float
            The start value of the sweep.
        stop : float
            The stop value of the sweep.
        
        Returns:
        --------
        ParameterSweep
            The parameter sweep object.
        """

        self._param_buffer.append((start,None,stop))
        return self
    
    def step(self, start: float, stepsize: float) -> ParameterSweep:
        """Adds a stepped sweep of values to the parameter sweep.
        
        Parameters:
        -----------
        start : float
            The start value of the sweep.
        stepsize : float
            The step size of the sweep.
        
        Returns:
        --------
        ParameterSweep
            The parameter sweep object.
        """
        self._param_buffer.append((start,stepsize,None))

    def add(self, N: int) -> tuple[Param] | Param:
        """Adds a dimension to the parameter sweep.
        
        Parameters:
        -----------
        N : int
            The number of steps in the dimension.
        
        Returns:
        --------
        tuple[Param]
            A tuple of Param objects."""
        params = []
        for start, step, stop in self._param_buffer:
            if step is None:
                params.append(Param.lin(start,stop,N))
            elif stop is None:
                params.append(Param.lin(start,start+step*N,N))
        self.add_dimension(*params)
        self._param_buffer = []
        if len(params) == 1:
            return params[0]
        return params
    
    def submit(self, S: np.ndarray, fs: np.ndarray) -> ParameterSweep:
        """Submits a set of S-parameters to the parameter sweep.
        
        Parameters:
        -----------
        S : np.ndarray
            The S-parameters to submit.
        """
        self._S_data.append((self._current_index, S))
        self._fdata = fs
        return self

    @property
    def S(self) -> NDSparameters:
        """Returns the S-parameters of the parameter sweep."""
        if self._mdim_data is None:
            self._mdim_data = self.generate_mdim()
        return self._mdim_data
    

    def generate_mdim(self) -> NDSparameters:
        """Generates a multi-dimensional array of S-parameters."""
        shape = [len(dimension[0]) for dimension in self.sweep_dimensions]
        s_proto = self._S_data[0][1]
        nports = s_proto.shape[0]
        ports = np.arange(1, nports+1)
        S = np.zeros(shape + list(self._S_data[0][1].shape), dtype=np.complex128)
        for index, s in self._S_data:
            S[index] = s._S
        NDS = NDSparameters(S, [np.array([dim._values for dim in dimension]) for dimension in self.sweep_dimensions] + [ports, ports, self._fdata])
        return NDS

    def iterate(self) -> Generator[tuple[tuple[int], tuple[float]], None, None]:
        '''An iterator that first compiles the total.'''
        self._S_data: list[np.ndarray] = []

        # Make a list of all dimensional index tuples as the product of the lengths of each dimension
        total = 1
        for dimension in self.sweep_dimensions:
            total *= len(dimension)

        # make a check for total above 10,000
        if total > 10000:
            raise ValueError(f"Total iterations ({total}) is above 10,000, are you sure you want to continue?")
        
        # Get all the length of the dimensions of the parameter sweep
        lengths = [len(dimension[0]) for dimension in self.sweep_dimensions]

        # make a list of indices like [(0,0,0),(1,0,0),(2,0,),...,(N,M,K)] using itertools
        indices = list(product(*[range(length) for length in lengths]))
        
        for ixs in indices:
            paramlist = []
            self._current_index: tuple[int] = ixs
            # set the index of each dimensional Param object
            for i, params in zip(ixs, self.sweep_dimensions):
                for param in params:
                    param.set_index(i)
                    param.initialize()
                    paramlist.append(param._value)
            yield ixs, tuple(paramlist)

    def add_dimension(self, *params: tuple[Param]):
        self.sweep_dimensions.append(params)

class MonteCarlo:
    """A class to represent a Monte Carlo simulation."""
    def __init__(self):
        self._random_numbers: list[Random] = []
        self.S: StatSparameters = None

    def gaussian(self, mean: float, std: float) -> Random:
        """Adds a Gaussian random number to the Monte Carlo simulation.
        
        Parameters:
        -----------
        mean : float
            The mean of the Gaussian distribution.
        std : float
            The standard deviation of the Gaussian distribution.
        
        Returns:
        --------
        Random
            A Random parameter object.
        """
        random = Random(lambda: np.random.normal(mean, std))
        random._mean = mean
        random._std = std
        random._value = mean
        self._random_numbers.append(random)
        return random
    
    def uniform(self, low: float, high: float) -> Random:
        """Adds a uniform random number to the Monte Carlo simulation.
        
        Parameters:
        -----------
        low : float
            The lower bound of the uniform distribution.
        high : float
            The upper bound of the uniform distribution.
        
        Returns:
        --------
        Random
            A Random parameter object.
        """
        random = Random(lambda: np.random.uniform(low, high))
        self._random_numbers.append(random)
        return random
    
    def iterate(self, N: int) -> Generator[int, None, None]:
        self.S = StatSparameters(N)
        for i in range(N):
            for random in self._random_numbers:
                random.initialize()
            yield i
    
    def submit_S(self, S: np.ndarray) -> None:
        """Submits S-parameters to the Monte Carlo simulation."""
        self.S.add(S)



def enforce_simparam(value: float | SimParam | Callable, inverse: bool = False, unit: str = '') -> SimParam:
    """Parses a numeric value to a SimParam object.

    Parameters:
    -----------
    value : float | Scalar | Callable
        The value to parse.
    inverse : bool
        Whether to return the inverse of the value.
    
    Returns:
    --------
    SimParam
        The SimParam object.
    """
    if isinstance(value, SimParam):
        if inverse:
            return value.inverse()
        return value
    elif isinstance(value, Callable):
        if inverse:
            return Function(lambda f: 1/value(f))
        return Function(value, unit=unit)
    elif isinstance(value, (int, float, complex)):
        if inverse:
            return Scalar(1/value, unit=unit)
        return Scalar(value, unit=unit)
    else:
        raise ValueError(f"Invalid value type: {type(value)}")

def enforce_value(value: float | SimParam) -> float | complex:
    if isinstance(value, SimParam):
        return value.value
    elif isinstance(value, Callable):
        return value(0)
    else:
        return float(value)  

def ensure_simparam(func):
    # Get the function signature
    sig = inspect.signature(func)

    @wraps(func)
    def wrapper(*args, **kwargs):
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()

        # Convert arguments to float if their type hint is complex or float
        for name, value in bound_args.arguments.items():
            param = sig.parameters[name]
            if param.annotation in (float, SimParam, float | SimParam, SimParam | float):
                bound_args.arguments[name] = enforce_simparam(value)

        return func(*bound_args.args, **bound_args.kwargs)
    return wrapper

def set_print_frequency(frequency: float) -> None:
    """Sets the frequency at which the simulation parameters are evaluated for printing.
    
    Parameters:
    -----------
    frequency : float
        The frequency at which the simulation parameters are evaluated.
    """
    # check if frequency is a float with a valid value
    if not isinstance(frequency, (int, float)):
        raise ValueError("Frequency must be a float")
    
    # check if it is greater than 0 and less than the upper frequency of the optical region
    if 0 < frequency < 1e15:
        raise ValueError("Frequency must be greater than 0 and less than 1e15 Hz")
    
    SimParam._eval_f = frequency


"""
Parameter Sweeps
In this demo we will quickly discuss how to do Paramter sweeps in heavi. It has a unique system to simplify the process.

"""

# We import Heavi
import heavi as hv

M = hv.Model()

# To execute a parameter sweep we creat the ParameterSweep class. 
sweep = hv.ParameterSweep()

# We can create new sweep axes using the lin method for example.
# First we call .lin(start,stop) to specify a value range. Then we call .add(N) to say: Create N points from start to stop.
Cs = sweep.lin(1e-12,2e-12).add(20)

# Sometimes we want to create two parameters that oth will change simultaneously instead of only one. Thus creating pairs. We can do this by repeatedly calling .lin()
# The .add(N) method will then return each linked parameters.
Rs, Ls = sweep.lin(50,100).lin(1e-9,5e-9).add(10)


# Cs, Rs and Ls are of type Param which is a sort of value that doesn't exist yet.

# Lets create our nodes and define our simple network.
nin = M.quick_port(50)
nout = M.quick_port(50)
nmid = M.node()

# The network is quite random but as you see we plug in the Param objects into our resistor. In this case our Resistor is now not defined by a single value
# but by a parameter that may have its value changed.
M.resistor(nin,nmid,Rs)
M.inductor(nmid,nout,Ls)
M.capacitor(M.gnd,nmid,Cs)

# To simmualte we define our frequency range
f = hv.frange(1e9,2e9,1001)

# We want to iterate over our entire parameter space so we call the iterate() method of our sweep function.
# i will be the nth iteration
for i in sweep.iterate():
    # We can compute the S-parameters and plug them into our sweep object together with the frequency.
    Sparameters = M.run_sp(f)
    sweep.submit(Sparameters,f)

# Finally we can call the MdimSparam function. This will create a multi-dimensional array of in this case shape 20,10,nF due to the order in which we 
# added our sweep axes.

MdimSparam = sweep.generate_mdim()

# We plot all our parameters superimposed. Its not interesting to look at but it serves its purpose.
hv.plot_s_parameters(f, [MdimSparam.S21[i,j,:] for i in range(20) for j in range(10)])


########################################################################################
##
##    The model module contains the Model and QuickModel classes.
##    An extended version of the Network class, the Model class is used to create
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


from .network import Network, Node
from .filtering import FilterType, BandType, CauerType, Filtering

#  __             __   __      __   ___  ___          ___    __        __  
# /  ` |     /\  /__` /__`    |  \ |__  |__  | |\ | |  |  | /  \ |\ | /__` 
# \__, |___ /~~\ .__/ .__/    |__/ |___ |    | | \| |  |  | \__/ | \| .__/ 
# -------------------------------------------


class Model(Network):
    """The Model class is an extension of the Network class. It can be used instead of the Network
    class and offers additional functionality for creating circuits.
    """    
    def __init__(self, default_name: str = "Node", 
                 filter_library: Filtering = Filtering, 
                 suppress_loadbar: bool = False):
        """Initializes the Model object.

        Args:
            default_name (str, optional): Default node name prefix. Defaults to "Node".
            filter_library (Filtering, optional): The filter library to use. Defaults to Filtering.
            suppress_loadbar (bool, optional): Whether or not to show numba loadbars. Defaults to False.
        """        
        super().__init__(default_name, suppress_loadbar=suppress_loadbar)
        self.filters: Filtering = filter_library(self)
        self.numbered_nodes: dict[int, Node] = dict()

    def __call__(self, index: int) -> Node:
        if index not in self.numbered_nodes:
            self.numbered_nodes[index] = self.node()
        node = self.numbered_nodes[index]
        node._parent = self
        return node
    
    


from .network import Network
from .node import Node
from emsutil import plot, smith, plot_sp
from .sparam import Sparameters, frange, NDSparameters, StatSparam, StatSparameters
from .model import Model
from .filtering import FilterType, BandType, CauerType, Filtering
from .numeric import MonteCarlo, Param, Function, ParameterSweep, set_print_frequency
from .transformations import VSWR_to_S11, S11_to_Z, S11_to_VSWR, Z_to_S11
from .lib.touchstone import FileBasedNPort

PI = 3.141592653589793
plot_s_parameters = plot_sp


########################################################################################
##
##    Stochastic Analysis of S-parameters
##    
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


import numpy as np
import matplotlib.pyplot as plt

#  ___            __  ___    __        __  
# |__  |  | |\ | /  `  |  | /  \ |\ | /__` 
# |    \__/ | \| \__,  |  | \__/ | \| .__/ 
# -------------------------------------------


def analyse_sparameter(sparam: list[np.ndarray], sparam_phase: list[np.ndarray], phase_reference: np.ndarray=None, n_bins: int = 51) -> tuple[tuple[float, float], tuple[float, float]]:
    """
    Analyse a set of S-parameter arrays, returning amplitude (dB) and phase statistics
    and plotting their histograms (PDF) and cumulative distributions (CDF), each on a
    separate y-axis.

    Parameters
    ----------
    sparam : list of np.ndarray
        Each element is a complex-valued S-parameter array (e.g. shape (F,)).
        We flatten them over frequency.
    *args, **kwargs : 
        Additional arguments (unused here, but provided for flexibility).

    Returns
    -------
    (amp_mean, amp_std), (phase_mean, phase_std) : tuple of tuples
        - amp_mean : float
            Mean of the magnitude in dB (flattened across frequency and Monte Carlo runs).
        - amp_std : float
            Standard deviation of the magnitude in dB.
        - phase_mean : float
            Mean of the phase error (degrees) after removing the mean complex phase.
        - phase_std : float
            Standard deviation of the phase error (degrees).
    """

    if sparam_phase is None:
        sparam_phase = sparam

    # ------------------
    # 1) Flatten S-parameters across all samples and frequencies
    # ------------------
    sparams_flat = np.concatenate(sparam)  # shape: (N*F,)
    
    # ------------------
    # 2) Compute amplitude in dB
    # ------------------
    magnitudes = np.abs(sparams_flat)
    magnitudes_db = 20.0 * np.log10(magnitudes + 1e-20)  # add small offset to avoid log(0)

    amp_mean = np.mean(magnitudes_db)
    amp_std = np.std(magnitudes_db)

    # ------------------
    # 3) Compute phase relative to average phase
    # ------------------
    # "Average complex S-parameter" means the mean of the complex samples.
    # NOTE: your code used: avg_sparam = sum(sparam)/len(sparam)
    # but that is effectively the average array. Then you took angles of ratio (S / avg_sparam).
    # The expression below just replicates that logic, but flattened. 
    # If you prefer the EXACT same logic as your snippet, do it the same way:
    if phase_reference is not None:
        phase_error_deg = 180 / np.pi * np.concatenate([
            np.angle(S / phase_reference) for S in sparam_phase
        ])
    else:
        phase_error_deg = 180 / np.pi * np.angle(np.concatenate(sparam_phase))
    
    phase_mean = np.mean(phase_error_deg)
    phase_std = np.std(phase_error_deg)

    # ------------------
    # 4) Plot results (PDF & CDF on separate y-axes)
    # ------------------
    fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12, 5))
    ax_amp, ax_phase = axes

    # Number of bins
    # === Amplitude Plot ===
    # PDF
    pdf_vals, bin_edges = np.histogram(magnitudes_db, bins=n_bins, density=True)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
    # CDF (numerical integration of the PDF)
    cdf_vals = np.cumsum(pdf_vals) * np.diff(bin_edges)[0]
    
    # Create a twin axis for the CDF
    ax_amp_pdf = ax_amp
    ax_amp_cdf = ax_amp_pdf.twinx()

    pdf_line = ax_amp_pdf.plot(bin_centers, pdf_vals, 'b', label='PDF')
    cdf_line = ax_amp_cdf.plot(bin_centers, cdf_vals, 'r', label='CDF')

    ax_amp_pdf.set_title(f"Amplitude Distribution\nMean={amp_mean:.2f} dB, Std={amp_std:.2f} dB")
    ax_amp_pdf.set_xlabel("Amplitude [dB]")
    ax_amp_pdf.set_ylabel("PDF", color='b')
    ax_amp_cdf.set_ylabel("CDF", color='r')
    ax_amp_pdf.grid(True)
    ax_amp_cdf.grid(False)

    # Combine legends
    lines = pdf_line + cdf_line
    labels = [l.get_label() for l in lines]
    ax_amp_cdf.legend(lines, labels, loc='best')

    # === Phase Plot ===
    # PDF
    pdf_vals, bin_edges = np.histogram(phase_error_deg, bins=n_bins, density=True)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
    # CDF
    cdf_vals = np.cumsum(pdf_vals) * np.diff(bin_edges)[0]

    # Create a twin axis for the CDF
    ax_phase_pdf = ax_phase
    ax_phase_cdf = ax_phase_pdf.twinx()

    pdf_line = ax_phase_pdf.plot(bin_centers, pdf_vals, 'b', label='PDF')
    cdf_line = ax_phase_cdf.plot(bin_centers, cdf_vals, 'r', label='CDF')

    ax_phase_pdf.set_title(f"Phase Error Distribution\nMean={phase_mean:.2f}°, Std={phase_std:.2f}°")
    ax_phase_pdf.set_xlabel("Phase Error [degrees]")
    ax_phase_pdf.set_ylabel("PDF", color='b')
    ax_phase_cdf.set_ylabel("CDF", color='r')
    ax_phase_pdf.grid(True)
    ax_phase_cdf.grid(False)

    # Combine legends
    lines = pdf_line + cdf_line
    labels = [l.get_label() for l in lines]
    ax_phase_cdf.legend(lines, labels, loc='best')

    plt.tight_layout()
    plt.show()

    # ------------------
    # 5) Return the stats
    # ------------------
    return (amp_mean, amp_std), (phase_mean, phase_std)


    


########################################################################################
##
##    Filtering Functions
##    This module contains functions to design filters and impedance transformers.
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------

from .network import Network, Node
from enum import Enum
import sympy as sym
from sympy.simplify.fu import TR0,TR8
import numpy as np
from loguru import logger


#  __   __             ___         ___       __   ___     ___            __  ___    __        __  
# /  ` /  \ |\ | \  / |__  |\ | | |__  |\ | /  ` |__     |__  |  | |\ | /  `  |  | /  \ |\ | /__` 
# \__, \__/ | \|  \/  |___ | \| | |___ | \| \__, |___    |    \__/ | \| \__,  |  | \__/ | \| .__/ 
# -------------------------------------------


def even(x):
    """ Returns True if x is even, False otherwise. """
    if x % 2 == 0:
        return True
    return False

def odd(x):
    """ Returns True if x is odd, False otherwise. """
    if x % 2 == 1:
        return True
    return False

#  __             __   __   ___  __  
# /  ` |     /\  /__` /__` |__  /__` 
# \__, |___ /~~\ .__/ .__/ |___ .__/ 
# -------------------------------------------


class FilterType(Enum):
    """Filter type enumeration.
    This enumeration is used to specify the type of filter to create.
    
    constants:
    -----------
    CHEBYCHEV : int
        Chebyshev filter.
    BUTTERWORTH : int
    """
    CHEBYCHEV = 1
    BUTTERWORTH = 2


class BandType(Enum):
    """Band type enumeration.
    This enumeration is used to specify the type of band to create.
    
    constants:
    -----------
    HIGHPASS : int
        Highpass filter.
    LOWPASS : int
        Lowpass filter.
    BANDPASS : int
        Bandpass filter.
    BANDSTOP : int
        Bandstop filter.
    """
    HIGHPASS = 1
    LOWPASS = 2
    BANDPASS = 3
    BANDSTOP = 4


class CauerType(Enum):
    """Cauer type enumeration.

    constants:
    -----------
    TYPE1 : int
        Type 1 Passband ripple
    TYPE2 : int
        Type 2 Stopband ripple

    """
    TYPE1 = 1
    TYPE2 = 2


def prototype_chebychev(order: int, ripple: float) -> tuple[np.ndarray, float]:
    """Generates the Chebyshev prototype filter G-coefficients.

    Parameters
    ----------
    order : int
        The order of the filter.
    ripple : float
        The ripple factor in dB.

    Returns
    -------
    tuple
        A tuple containing the G-coefficients and the load impedance scaler
    
    """
    N = order
    
    Gk = [i for i in range(N + 2)]
    Ak = [i for i in range(N + 2)]
    Bk = [i for i in range(N + 2)]

    B = np.log(1 / np.tanh(ripple * np.log(10) / 40))
    y = np.sinh(B / (2 * N))

    for k in Gk:
        Ak[k] = np.sin((2 * k - 1) * np.pi / (2 * N))
        Bk[k] = y**2 + np.sin(k * np.pi / N) ** 2
        if k == 0:
            Gk[k] = 1
        elif k == 1:
            Gk[k] = 2 * Ak[1] / y
        elif k <= N:
            Gk[k] = 4 * Ak[k - 1] * Ak[k] / (Bk[k - 1] * Gk[k - 1])
        elif k == N + 1:
            if k % 2 == 0:
                Gk[k] = 1
            else:
                Gk[k] = 1 / np.tanh(B / 4) ** 2
    return Gk[1:-1], Gk[-1]


def prototype_butterworth(order: int) -> tuple[np.ndarray, float]:
    """Generates the Butterworth prototype filter G-coefficients.
    
    Parameters
    ----------
    order : int
        The order of the filter.
    
    Returns
    -------
    tuple
        A tuple containing the G-coefficients and the load impedance scaler (for generality, the scaler is always 1)
    
    """

    N = order
    Gk = [i + 1 for i in range(N)]

    for i in Gk:
        Gk[i - 1] = 2 * np.sin(np.pi * (2 * i - 1) / (2 * N))

    return Gk, 1


def gen_cheb_poly(n: int):
    """Generates the nth order Chebyshev polynomial."""
    x = sym.symbols('x')
    if n==0:
        output = 1
    elif n==1:
        T1 = x
        output = T1
    else:
        Tnm1 = x
        Tnm2 = 1
        Tn = 1
        for n in range(n-1):
            Tn = 2*x*Tnm1-Tnm2
            Tnm2 = Tnm1
            Tnm1 = Tn
        output = Tn
    theta, secm = sym.symbols('theta secm')
    output = sym.expand(output,x).subs(x,sym.cos(theta)*secm)
    for i in range(n):
        output = TR0(sym.expand(TR8(output)))
    return sym.collect(output, theta, exact=None)

def impedance_transformer_cheb(Z1: float, Z2: float, A: float, N: int):
    """Generates the impedances for an impedance transformer using Chebyshev polynomials.
    
    Parameters:
    -----------
    Z1 : float
        The characteristic impedance of the first port.
    Z2 : float
        The characteristic impedance of the second port.
    A : float
        The ripple factor in dB.
    N : int
        The number of sections in the transformer.

    """
    theta = sym.symbols('theta')
    func = 2
    fout = 0
    i = 0
    for n in range(N//2+1):
        Gs = sym.symbols(f'Gamma{i}')
        fout = fout + func*Gs*sym.cos((N-n*2)*theta)
        i = i+1
    
    cheb_poly = gen_cheb_poly(N)
    gammas = []
    cheb_poly_output = cheb_poly
    gamma_poly_output = fout
    for i in range(0,N//2+1):
        mult = N-i*2
        symbol = sym.Symbol(f'Gamma{i}')
        T1 = fout.coeff(sym.cos(mult*theta))
        T2 = A*cheb_poly.coeff(sym.cos(mult*theta))
        if mult==0:
            T1 = gamma_poly_output
            T2 = A*cheb_poly_output*2
        gamma = sym.solve(T1-T2, symbol)[0]
        secmval = np.cosh(1/N*np.arccosh((1/(2*A))*np.log(Z2/Z1)))
        gamma_val = gamma.subs(sym.Symbol('secm'), secmval)
        gammas.append(gamma_val)
        cheb_poly_output = cheb_poly_output - cheb_poly.coeff(sym.cos(mult*theta))*sym.cos(mult*theta)
        gamma_poly_output = gamma_poly_output - fout.coeff(sym.cos(mult*theta))*sym.cos(mult*theta)
    gammas = gammas + list(reversed(gammas))[((N+1)%2):]
    impedances = [Z1,]
    Zlast = Z1
    for g in gammas[:-1]:
        Zsec = Zlast*np.exp(2.0*float(g))
        impedances.append(Zsec)
        Zlast = Zsec
    impedances.append(Z2)
    return impedances


class Filtering:

    def __init__(self, N: Network):
        self.N = N 

    def impedance_transformer(self,
                              gnd: Node,
                              port1: Node,
                              port2: Node,
                              Z01: float,
                              Z02: float,
                              fc: float = None,
                              Nsections: int = 2,
                              ripple: float = 0.05,
                              er: float = 1,
                              frange: tuple[float, float] = None
                              ):
        """
        Designs an impedance transformer using Chebyshev polynomials.
        Parameters:
        -----------
        gnd : Node
            The ground node.
        port1 : Node
            The first port node.
        port2 : Node
            The second port node.
        Z01 : float
            The characteristic impedance of the first port.
        Z02 : float
            The characteristic impedance of the second port.
        fc : float, optional
            The center frequency of the transformer. If `frange` is provided, `fc` is computed as the center of `frange`.
        Nsections : int, optional
            The number of sections in the transformer. Default is 2.
        ripple : float, optional
            The ripple factor for the Chebyshev polynomial. Default is 0.05.
        er : float, optional
            The relative permittivity of the medium. Default is 1.
        frange : tuple[float, float], optional
            The frequency range for the transformer. If provided, `fc` and `ripple` are computed based on this range.
        Returns:
        --------
        tuple
            A tuple containing the list of impedances for each section and the computed ripple factor.
        Notes:
        ------
        This function uses Chebyshev polynomials to design an impedance transformer with the specified parameters.
        """
        
        if frange is not None:
            f1, f2 = frange
            fc = 0.5*(f1+f2)
            A = np.cosh(Nsections*np.arccosh(1/np.cos(np.pi/4*(2-(f2-f1)/(fc)))))
            ripple = 1/(2*A) * np.log(Z02/Z01)
            logger.debug(f'Computed Ripple = {ripple}')

        l0 = 299792458/fc * 0.25 * (1/np.sqrt(er))
        imps = impedance_transformer_cheb(Z01, Z02, ripple, Nsections)[1:-1]
        nodes = [self.N.node(index='TransformerNode') for i in range(len(imps)-1)]

        nodes = [port1, ] + nodes + [port2,]

        for Z, n1, n2 in zip(imps, nodes[:-1], nodes[1:]):
            logger.debug(f'Transmisison line from {n1} to {n2} with Z0={Z}')
            self.N.transmissionline(gnd, n1, n2, Z, er, l0)
        return imps, ripple
    
    def cauer_filter(
        self,
        gnd: Node,
        port1: Node,
        port2: Node,
        fc: float,
        bw: float,
        order: int = 2,
        ripple: float = 0.05,
        kind: FilterType = FilterType.BUTTERWORTH,
        type: BandType = BandType.LOWPASS,
        Z0: float = 50,
        cauer_type: CauerType = CauerType.TYPE2,
        chebychev_correction: bool = False,
    ):
        """Creates a Cauer ladder filter.
        This function will create a Cauer ladder filter with the specified parameters between port1 and port2.

        Parameters
        ----------
        gnd : Node
            The ground node.
        port1 : Node
            The input port node.
        port2 : Node
            The output port node.
        fc : float
            The cutoff frequency of the filter.
        bw : float
            The bandwidth of the filter.
        order : int, optional
            The order of the filter, by default 2.
        ripple : float, optional
            The ripple of the filter, by default 0.05.
        kind : FilterType, optional
            The type of filter, by default FilterType.BUTTERWORTH.
        type : BandType, optional
            The type of band, by default BandType.LOWPASS.
        Z0 : float, optional
            The reference impedance, by default 50.
        cauer_type : CauerType, optional
            The type of Cauer filter, by default CauerType.TYPE2.
        chebychev_correction : bool, optional
            Whether to apply Chebychev correction, by default False. If turned to true, the cutoff frequency and bandwidth will be corrected to match the ripple.
        
        """

        NODEINDEX = "internalFilterNode"
        cauernodes = [port1]
        Nint = 0
        if cauer_type == CauerType.TYPE2:
            Nint = np.floor((order - 1) / 2)
        elif cauer_type == CauerType.TYPE1:
            Nint = np.floor((order - 2) / 2)

        Nint = int(Nint)
        for i in range(0, Nint):
            cauernodes.append(self.N.node(index=NODEINDEX))
        cauernodes.append(port2)

        Capacitors = []
        Inductors = []

        w_cutoff = 2 * np.pi * fc
        w0_bandpass = w_cutoff
        delta_omega = 2 * np.pi * bw

        Qfactor = fc / bw

        LoadImpedanceScaler = 1

        if kind == FilterType.BUTTERWORTH:
            Gk, LoadImpedanceScaler = prototype_butterworth(order)
        elif kind == FilterType.CHEBYCHEV:
            Gk, LoadImpedanceScaler = prototype_chebychev(order, ripple)
            if chebychev_correction:
                epsilon = np.sqrt(10 ** (ripple / 10) - 1)

                if type is BandType.HIGHPASS or type is BandType.BANDSTOP:
                    delta_omega = delta_omega * np.cosh(
                        1 / order * np.arccosh(1 / epsilon)
                    )
                    w_cutoff = w_cutoff * np.cosh(1 / order * np.arccosh(1 / epsilon))
                else:
                    delta_omega = delta_omega / np.cosh(
                        1 / order * np.arccosh(1 / epsilon)
                    )
                    w_cutoff = w_cutoff / np.cosh(1 / order * np.arccosh(1 / epsilon))
                Qfactor = w0_bandpass / delta_omega

        add_counter = 0

        if cauer_type == CauerType.TYPE1:
            testf = odd
            LoadImpedanceScaler = 1 / LoadImpedanceScaler
        elif cauer_type == CauerType.TYPE2:
            testf = even
            add_counter = 1

        for i, gk in enumerate(Gk):
            k = i + 1
            node1 = int(np.floor((i + add_counter) / 2))
            node2 = node1 + 1
            if type == BandType.LOWPASS:
                if testf(k):
                    C = self.N.capacitor(gnd, cauernodes[node1], gk / (w_cutoff * Z0))
                    Capacitors.append(C)
                else:
                    L = self.N.inductor(
                        cauernodes[node1], cauernodes[node2], (gk * Z0) / w_cutoff
                    )
                    Inductors.append(L)
            elif type == BandType.HIGHPASS:
                if testf(k):
                    L = self.N.inductor(gnd, cauernodes[node1], Z0 / (gk * w_cutoff))
                    Inductors.append(L)
                else:
                    C = self.N.capacitor(
                        cauernodes[node1], cauernodes[node2], 1 / (w_cutoff * gk * Z0)
                    )
                    Capacitors.append(C)
            elif type == BandType.BANDPASS:
                if testf(k):
                    Cpar = Qfactor * gk / (w0_bandpass * Z0)
                    Lpar = Z0 / (gk * w0_bandpass * Qfactor)
                    Cp = self.N.capacitor(gnd, cauernodes[node1], Cpar)
                    Lp = self.N.inductor(gnd, cauernodes[node1], Lpar)
                    Capacitors.append(Cp)
                    Inductors.append(Lp)
                else:
                    xnode = self.N.node(index=NODEINDEX)
                    Cseries = 1 / (w0_bandpass * gk * Z0 * Qfactor)
                    Lseries = Qfactor * gk * Z0 / w0_bandpass
                    Cser = self.N.capacitor(cauernodes[node1], xnode, Cseries)
                    Lser = self.N.inductor(xnode, cauernodes[node2], Lseries)
                    Capacitors.append(Cser)
                    Inductors.append(Lser)
            elif type is BandType.BANDSTOP:
                if testf(k):
                    xnode = self.N.node(index=NODEINDEX)
                    Cpar = gk / (Qfactor * w0_bandpass * Z0)
                    Lpar = Qfactor * Z0 / (gk * w0_bandpass)
                    Cp = self.N.capacitor(gnd, xnode, Cpar)
                    Lp = self.N.inductor(xnode, cauernodes[node1], Lpar)
                    Capacitors.append(Cp)
                    Inductors.append(Lp)
                else:
                    Cseries = Qfactor / (w0_bandpass * gk * Z0)
                    Lseries = gk * Z0 / (w0_bandpass * Qfactor)
                    Cser = self.N.capacitor(cauernodes[node1], cauernodes[node2], Cseries)
                    Lser = self.N.inductor(cauernodes[node1], cauernodes[node2], Lseries)
                    Capacitors.append(Cser)
                    Inductors.append(Lser)
        return Capacitors, Inductors, LoadImpedanceScaler
    



########################################################################################
##
##    S-parameter optimization features
##    Features for optimizing S-parameters of a network.
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


from ..numeric import SimParam
from ..network import Network
from ..sparam import Sparameters
from typing import Callable
import numpy as np
from enum import Enum
from abc import ABC, abstractmethod

#  __             __   __   ___  __  
# /  ` |     /\  /__` /__` |__  /__` 
# \__, |___ /~~\ .__/ .__/ |___ .__/ 
# -------------------------------------------


class OptimVar(SimParam):
    """ Class for defining an optimisation variable. 
    
    This class is used to define a variable that can be optimised. It is a subclass of SimParam, and can be used in the same way.
    
    Parameters
    ----------
    initial_value : float
        The initial value of the variable.
    bounds : tuple[float, float], optional
        The bounds of the variable. If not specified, the variable is unbounded.
    mapping : Callable, optional
        A function that maps the variable to a different space. If not specified, the variable is used as
        is. The function should take a float and return a float.
    """
    def __init__(self, initial_value: float, 
                 bounds: tuple[float, float] | None = None, 
                 mapping: Callable | None = None,
                 unit: str = ''):
        if not isinstance(mapping, Callable) and mapping is not None:
            raise TypeError(f'The variable mapping must be a callable function. Use dBbelow or dBabove. Not {type(mapping)}')
        self._value = initial_value
        self.bounds = bounds
        self.mapping: Callable = mapping
        self.unit: str = unit

    def set(self, value: float):
        self._value = value

    def __call__(self, f: np.ndarray) -> np.ndarray:
        if self.mapping is None:
            return self._value*np.ones_like(f)
        else:
            return self.mapping(self._value)*np.ones_like(f)

class PNorm(Enum):
    """ Enum for defining the type of norm to use in the optimisation.
    
    This enum is used to define the type of norm to use in the optimisation. It can be used to specify the type of norm to use in the optimisation.
    
    Attributes
    ----------
    MIN : int
        The minimum norm.
    MAX : int
        The maximum norm.
    P1 : int
        The P1 norm (average).
    P2 : int
        The P2 norm (root mean square).
    """
    MIN = 0
    MAX = 1
    P1 = 2
    P2 =3

    def get_metric(self) -> Callable:
        """ Get the metric function for the norm."""
        if self is self.MAX:
            return np.max
        if self is self.MIN:
            return np.min
        if self is self.P1:
            return np.mean
        if self is self.P2:
            return lambda x: np.sqrt(np.mean(x**2))

def generalized_mean(norm: float) -> Callable:
    """ Generate a generalized mean function.
    
    This function generates a generalized mean function with the specified norm.
    
    Parameters
    ----------
    norm : float
        The norm of the mean.
    
    Returns
    -------
    Callable
        The generalized mean function.
        """
    def func(x):
        return np.mean(x**norm)**(1/norm)
    return func


def dBbelow(dB_level: float,
            norm: float | PNorm = PNorm.P2) -> Callable:
    """ Generate a function that calculates the dB level below a specified level.
    
    This function generates a function that calculates the dB level below a specified level.
    
    Parameters
    ----------
    dB_level : float
        The dB level to calculate the level below.
    norm : float | PNorm, optional
        The norm to use in the calculation. If not specified, the P2 norm is used.
    
    Returns
    -------
    Callable
        The function that calculates the dB level below the specified level.
    """

    if isinstance(norm, PNorm):
        meval = norm.get_metric()
    else:
        meval = generalized_mean(norm)
    
    def metric(S: np.ndarray) -> float:
        return meval(np.clip(20*np.log10(np.abs(S))-dB_level, a_min=0, a_max=None))/10
    return metric

def dBabove(dB_level: float,
            norm: float | PNorm = PNorm.P2) -> Callable:
    """ Generate a function that calculates the dB level above a specified level.
    
    This function generates a function that calculates the dB level above a specified level.
    
    Parameters
    ----------
    dB_level : float
        The dB level to calculate the level above.
    norm : float | PNorm, optional
        The norm to use in the calculation. If not specified, the P2 norm is used.
    
    Returns
    -------
    Callable
        The function that calculates the dB level above the specified level.
    """
    
    if isinstance(norm, PNorm):
        meval = norm.get_metric()
    else:
        meval = generalized_mean(norm)
    
    def metric(S: np.ndarray) -> float:
        return meval(np.clip(dB_level - 20*np.log10(np.abs(S)), a_min=0, a_max=None))/10
    
    return metric


class Requirement(ABC):
    """ Abstract class for defining a requirement.
    
    This class is used to define a requirement for an optimiser. It is an abstract class, and should be subclassed to define a specific requirement.
    
    """
    @abstractmethod
    def eval(self, S: Sparameters) -> float:
        pass


class FrequencyLevel(Requirement):

    def __init__(self, 
                 fmin: float,
                 fmax: float,
                 nF: int,
                 sparam: tuple[int, int],
                 upper_limit: float | None = None,
                 lower_limit: float | None = None,
                 norm: float = 2.0,
                 weight: float = 1,
                 f_norm: float | PNorm = 4):
        self.fs = np.linspace(fmin,fmax,nF)
        self.param: tuple[int, int] = sparam
        self.slc: slice = None
        self.weight: float = weight

        self.upper_limit_value: float | None = upper_limit
        self.lower_limit_value: float | None = lower_limit
        
        self.upper_limit: Callable | None = None
        self.lower_limit: Callable | None = None
        
        if upper_limit is not None and lower_limit is not None:
            self.upper_limit = dBbelow(self.upper_limit_value, norm)
            self.lower_limit = dBabove(self.lower_limit_value, norm)
            self.metric = lambda S: self.upper_limit(S) + self.lower_limit(S)
        elif upper_limit is not None:
            self.upper_limit = dBbelow(self.upper_limit_value, norm)
            self.metric = self.upper_limit
        elif lower_limit is not None:
            self.lower_limit = dBabove(self.lower_limit_value, norm)
            self.metric = self.lower_limit
        else:
            raise ValueError("At least one of upper_limit or lower_limit must be specified")
        

    def eval(self, S: Sparameters) -> float:
        value = self.metric(S.S(self.param[0],self.param[1])[self.slc])*self.weight
        return value
    
    def generate_fill_area(self) -> tuple[float, float, float, float]:
        lower = self.lower_limit_value
        upper = self.upper_limit_value
        if lower is None:
            lower = 0
        if upper is None:
            upper = -100
        return (self.fs[0], self.fs[-1], upper, lower)


class Optimiser:
    """ Class for defining an optimiser.
    
    This class is used to define an optimiser for a network. It is used to define the parameters to optimise, the requirements to meet, and the objective function to optimise.
    
    Parameters
    ----------
    network : Network
        The network to optimise.
    
    """
    def __init__(self, network: Network):
        self.network = network
        self.parameters: list[OptimVar] = []
        self.requirements: list[FrequencyLevel] = []

    @property
    def spec_area(self) -> list[tuple[float, float, float, float]]:
        """ Get the fill areas for the requirements."""
        print([req.generate_fill_area() for req in self.requirements])
        return [req.generate_fill_area() for req in self.requirements]
    
    @property
    def bounds(self) -> list[tuple[float, float]]:
        """ Get the bounds of the optimisation variables."""
        return [p.bounds for p in self.parameters]
    
    @property
    def x0(self) -> np.ndarray:
        """ Get the initial values of the optimisation variables."""
        return np.array([p.value for p in self. parameters])

    def add_param(self, initial, 
                  bounds: tuple[float, float] = None, 
                  mapping: Callable = None,
                  unit: str = '') -> OptimVar:
        """ Add an optimisation variable.
        
        This method adds an optimisation variable to the optimiser.
        
        Parameters
        ----------
        initial : float
            The initial value of the variable.
        bounds : tuple[float, float], optional
            The bounds of the variable. If not specified, the variable is unbounded.
        mapping : Callable, optional
            A function that maps the variable to a different space. If not specified, the variable is used as
            is. The function should take a float and return a float.
        
        Returns
        -------
        OptimVar
            The optimisation variable.
        """
        param = OptimVar(initial, bounds=bounds, mapping=mapping, unit=unit)
        self.parameters.append(param)
        return param
    
    def cap(self, 
            logscale: bool = True, 
            initial: float = -12,
            limits: tuple[float, float] = (-13, -6)) -> OptimVar:
        """ Add a capacitor parameter.
        
        This method adds a capacitor parameter to the optimiser.
        The default capacitor range is from 0.1 pF to 1 uF.
        
        Parameters
        ----------
        logscale : bool, optional
            Whether to use a log scale for the parameter. If not specified, a log scale is used.
        initial : float, optional
            The initial value of the parameter. If not specified, the default value is used (1 pF).
        limits : tuple[float, float], optional
            The limits of the parameter. If not specified, the default limits are used (0.1 pF to 1 uF).
        
        Returns
        -------
        OptimVar
            The capacitor parameter.
        """
        if not logscale:
            return self.add_param(10**initial, (1*10**limits[0], 1*10**limits[1]), unit='f')
        else:
            return self.add_param(initial, limits, mapping= lambda x: 10**(x), unit='f')
    
    def ind(self, 
            logscale: bool = True,
            initial: float = -9,
            limits: tuple[float, float] = (-12,-5)) -> OptimVar:
        """ Add an inductor parameter.
        
        This method adds an inductor parameter to the optimiser.
        
        Parameters
        ----------
        logscale : bool, optional
            Whether to use a log scale for the parameter. If not specified, a log scale is used.
        
        Returns
        -------
        OptimVar
            The inductor parameter.
        """
        if not logscale:
            return self.add_param(10**initial, (10**(limits[0]),10**(limits[1])), unit='H')
        else:
            return self.add_param(initial, limits, mapping= lambda x: 10**(x), unit='H')
    
    def add_requirement(self, req: Requirement):
        """ Add a requirement to the optimiser.
        
        This method adds a requirement to the optimiser.
        
        Parameters
        ----------
        req : Requirement
            The requirement to add.
        """
        self.requirements.append(req)

    def add_splimit(self,
                        fmin: float,
                        fmax: float,
                        nF: int,
                        sparam: tuple[int, int],
                        upper_limit: float | None = None,
                        lower_limit: float | None = None,
                        norm: float = 2.0,
                        weight: float = 1,
                        f_norm: float | PNorm = 4):
            """ Add a frequency level requirement to the optimiser.
            
            This method adds a frequency level requirement to the optimiser.
            
            Parameters
            ----------
            fmin : float
                The minimum frequency of the requirement.
            fmax : float
                The maximum frequency of the requirement.
            nF : int
                The number of frequencies to use in the requirement.
            sparam : tuple[int, int]
                The S-parameter to use in the requirement.
            upper_limit : Callable, optional
                The upper limit of the requirement. If not specified, the requirement is unbounded.
            lower_limit : Callable, optional
                The lower limit of the requirement. If not specified, the requirement is unbounded.
            weight : float, optional
                The weight of the requirement. If not specified, the weight is 1.
            f_norm : float | PNorm, optional
                The norm to use in the requirement. If not specified, the P2 norm is used.
            """
            req = FrequencyLevel(fmin, fmax, nF, sparam, upper_limit=upper_limit, lower_limit=lower_limit, norm=norm, weight=weight, f_norm=f_norm)
            self.add_requirement(req)

    def generate_objective(self, 
                           pnorm=2, 
                           initial: np.ndarray = None, 
                           differential_weighting: float = 0,
                           differential_weighting_exponent: float = 5) -> Callable:
        """ Generate the objective function for the optimiser.
        
        This method generates the objective function for the optimiser.
        
        Parameters
        ----------
        pnorm : int, optional
            The norm to use in the optimisation. If not specified, the P2 norm is used.
        initial : np.ndarray, optional
            The initial values of the optimisation variables. If not specified, the default values are used.
        differential_weighting : float, optional
            The differential weighting to use in the optimisation. If not specified, no differential weighting is used.
        differential_weighting_exponent : float, optional
            The exponent to use in the differential weighting. If not specified, the default value is used.
        
        Returns
        -------
        Callable
            The objective function for the optimiser.
        
        """
        if initial is not None:
            for p, v in zip(self.parameters, initial):
                p.value = v
        n = 0
        fs = []
        for req in self.requirements:
            fs = fs + list(req.fs)
            req.slc = slice(n,len(fs))
            n = len(fs)
        fs = np.array(fs)
        NR = len(self.requirements)
        if differential_weighting > 0:
            def objective(coeffs):
                for p,c in zip(self.parameters, coeffs):
                    p.set(c)
                S = self.network.run_sp(fs)
                Ms = np.array([abs(req.eval(S)) for req in self.requirements])
                return (np.mean(Ms**pnorm))**(1/pnorm) + differential_weighting*(max(Ms)-min(Ms))**differential_weighting_exponent
        else:
            def objective(coeffs):
                for p,c in zip(self.parameters, coeffs):
                    p.set(c)
                S = self.network.run_sp(fs)
                subobj = [abs(req.eval(S))**pnorm for req in self.requirements]
                return (sum(subobj)/NR)**(1/pnorm)
               
        return objective
    






"""
Optimization

In this example we will quickly show how to optimize your components in Heavi.

We are going to optimize a simple LC circuit to have the right values and cutoff frequency.
"""

import heavi as hv
from heavi.lib import optim

M = hv.Model()

p1 = M.quick_port(50)
p2 = M.quick_port(50)

# To optimize we create an optimizer object.
opt = optim.Optimiser(M)

# We create a new parameter but we want to distribute the range logarithmically.
# Thus we will optimize in the exponent range going from 10pH to 1uH etc. Then we apply a value mapping that computes the actual value.
Lopt = opt.add_param(-9, (-11,-6), mapping=lambda x: 10**x)
Copt = opt.add_param(-12,(-14,-9), mapping=lambda x: 10**x)

# We create our circuit.
M.inductor(p1,p2, Lopt,)
M.capacitor(M.gnd, p2, Copt)

# We are going to define some S-parameter constraint using the add_splimit function. The arguments are:
#  fmin: start frequency
#  fmax: ending frequency
#  nF: Number of frequency points to sample
#  Sp index: The Sparameter to study, in this case firs S11 and then S21
#  upper_limit: This is a funtion that should penalize certain value excurions. We can create a penalty term simply by using the optim.dBbelow function
#  weight: How important it is

# The optim.dBbelow(-15,norm=4) will create a penalty function that gets higher if the value is above the provided limit in dB's. The norm indicates the
# Lebesque norm: Thus, for each frequency f we have: Penalty = (sum penalty**n)**(1/n)

opt.add_splimit(0.1e9, 2e9, 21, (1,1), upper_limit=-15, norm=4, weight=5)
opt.add_splimit(10e9, 20e9, 21, (2,1), upper_limit=-15, norm=4)

# Heavi does not have a build in optimizer but we can simply use ones from Scipy! Differential evolution si very good
from scipy.optimize import differential_evolution

# We call the differential evolution on our objective function. We create this using the generate_objective() method
# Differential weighting is a feature to prevent optimizers from getting stuck at local minima by adding the differene between optimization variables as penalty term

objecte_function = opt.generate_objective(2,differential_weighting=True)

# Finally we can call the differential evolution function. The bounds for each parameter are simply available in the bounds property of our opt instance.
x0 = differential_evolution(objecte_function, bounds=opt.bounds, popsize=50)

print(x0)
M.print_components()

# After the final iteration has been called, that version of the parameters of our circuit are still online! W


fs = hv.frange(1e9,30e9,1001)

S = M.run_sp(fs)

hv.plot_s_parameters(fs,[S.S11,S.S21], labels=['S11','S21'], spec_area=opt.spec_area, logx=True)


########################################################################################
##
##    An S-parameter data container class
##    This module contains classes for handling S-parameter data
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


import numpy as np
from typing import Callable
import re
from scipy.interpolate import interpn, RegularGridInterpolator
from loguru import logger


#  ___            __  ___    __        __  
# |__  |  | |\ | /  `  |  | /  \ |\ | /__` 
# |    \__/ | \| \__,  |  | \__/ | \| .__/ 
# -------------------------------------------


def frange(fmin: float, fmax: float, n: int) -> np.ndarray:
    """Generates a linearly spaced frequency range

    Args:
        fmin (float): Lower frequency bound
        fmax (float): Upper frequency bound
        n (int): Number of points

    Returns:
        np.ndarray: Linearly spaced frequency range
    """    
    return np.linspace(fmin, fmax, n)


#  __             __   __   ___  __  
# /  ` |     /\  /__` /__` |__  /__` 
# \__, |___ /~~\ .__/ .__/ |___ .__/ 
# -------------------------------------------

class Sparameters:


    def __init__(self, S: np.ndarray, f: np.ndarray):
        """ S-parameters object
        
        Parameters
        ----------
        S : np.ndarray
            Scattering matrix of shape (nports, nports, nfreqs) 
        """
        self._S: np.ndarray = S
        self.f: float = f
        self.nports: int = S.shape[0]
        self.nfreqs: int = S.shape[2]
    
    @property
    def shape(self) -> tuple:
        return self._S.shape
    
    def S(self, p1: int, p2: int):
        """Get S-parameter S[p1, p2] counting from 1, S(1,1) is S11"""
        
        # check if p1 and p2 are valid ports
        if p1 < 1 or p1 > self.nports:
            raise ValueError(f"Port {p1} out of range")
        if p2 < 1 or p2 > self.nports:
            raise ValueError(f"Port {p2} out of range")
        return self._S[p1-1, p2-1, :]
    
    def __call__(self, p1: int, p2: int):
        return self.S(p1, p2)
    
    def __getattr__(self, name):
        match = re.match(r'^S(\d+)(\d+)$', name)
        if match:
            p1, p2 = int(match.group(1)), int(match.group(2))
            return self.S(p1, p2)
        else:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __dir__(self):
        # Return dynamically generated attributes for autocompletion
        valid_s_parameters = [f"S{m}{n}" for m in range(1, 6) for n in range(1, 5)]
        return valid_s_parameters + super().__dir__()
    

    ### All S-parameters of a 5 port network for autocompletion purposes.
    @property
    def S11(self):
        return self.S(1, 1)

    @property
    def S12(self):
        return self.S(1, 2)

    @property
    def S13(self):
        return self.S(1, 3)

    @property
    def S14(self):
        return self.S(1, 4)

    @property
    def S15(self):
        return self.S(1, 5)

    @property
    def S21(self):
        return self.S(2, 1)

    @property
    def S22(self):
        return self.S(2, 2)

    @property
    def S23(self):
        return self.S(2, 3)

    @property
    def S24(self):
        return self.S(2, 4)

    @property
    def S25(self):
        return self.S(2, 5)

    @property
    def S31(self):
        return self.S(3, 1)

    @property
    def S32(self):
        return self.S(3, 2)

    @property
    def S33(self):
        return self.S(3, 3)

    @property
    def S34(self):
        return self.S(3, 4)

    @property
    def S35(self):
        return self.S(3, 5)

    @property
    def S41(self):
        return self.S(4, 1)

    @property
    def S42(self):
        return self.S(4, 2)

    @property
    def S43(self):
        return self.S(4, 3)

    @property
    def S44(self):
        return self.S(4, 4)

    @property
    def S45(self):
        return self.S(4, 5)

    @property
    def S51(self):
        return self.S(5, 1)

    @property
    def S52(self):
        return self.S(5, 2)

    @property
    def S53(self):
        return self.S(5, 3)

    @property
    def S54(self):
        return self.S(5, 4)

    @property
    def S55(self):
        return self.S(5, 5)


class Interpolator:
    """Wrapper class for scipy RegularGridInterpolator
    """    
    def __init__(self, interp: RegularGridInterpolator):
        self._interp: RegularGridInterpolator = interp

    def __call__(self, *args) -> np.ndarray:
        """Interpolates the data at the given points

        Args:
            *args: N-dimensional points to interpolate at
        
        Returns:
            np.ndarray: Interpolated data
        """        
        return np.squeeze(self._interp(np.meshgrid(*args, indexing='ij')))

class StatSparam:

    def __init__(self, S: np.ndarray):
        self._S: np.ndarray = S
        self.ntrials: int = S.shape[0]

    @property
    def amplitude_mean(self) -> np.ndarray:
        return np.abs(self._S).mean(axis=0)
    
    @property
    def amplitude_std(self) -> np.ndarray:
        return np.abs(self._S).std(axis=0)
    
    @property
    def power_mean(self) -> np.ndarray:
        return (np.abs(self._S)**2.0).mean(axis=0)
    
    @property
    def power_std(self) -> np.ndarray:
        return (np.abs(self._S)**2.0).std(axis=0)
    
    @property
    def rms_mean(self) -> np.ndarray:
        return np.sqrt(np.abs(self._S)**2.0).mean(axis=0) 
    
    @property
    def rms_std(self) -> np.ndarray:
        return np.sqrt(np.abs(self._S)**2.0).std(axis=0)


class StatSparameters(Sparameters):
    """Series of S-parameters over multiple Monte Carlo runs
    """

    def __init__(self, Ntrials: int):
        self.Ntrials: int = Ntrials
        self._Sdata: list[np.ndarray] = []

    @property
    def nports(self) -> int:
        if len(self._Sdata) == 0:
            return 0
        return self._Sdata[0].shape[0]
    
    @property
    def nfreqs(self) -> int:
        if len(self._Sdata) == 0:
            return 0
        return self._Sdata[0].shape[2]
    
    def add(self, S: np.ndarray | Sparameters):
        """Add S-parameter data to the container

        Args:
            S (np.ndarray): S-parameter data of shape (nports, nports, nfreqs)
        """     
        if isinstance(S, Sparameters):
            S = S._S
        
        if not self._Sdata:
            self._Sdata.append(S)
            return   
        if S.shape != (self.nports, self.nports, self.nfreqs):
            raise ValueError(f"Invalid S-parameter shape. Expected {(self.nports, self.nports, self.nfreqs)}, got {S.shape}")
        self._Sdata.append(S)
    
    def S(self, p1: int, p2: int) -> StatSparam:
        """Get S-parameter S[p1, p2] counting from 1, S(1,1) is S11

        Args:
            p1 (int): Port number
            p2 (int): Port number
        
        Returns:
            np.ndarray: S-parameter data
        """        
        return StatSparam(np.stack([S[p1-1, p2-1, :] for S in self._Sdata], axis=0))
    
    def __call__(self, p1: int, p2: int) -> StatSparam:
        return self.S(p1, p2)
    
    

class NDSparameters(Sparameters):
    """N-dimensional S-parameter data container class
    """

    def __init__(self, mdim_data: np.ndarray, axes: list[np.ndarray]):
        """Initializes the N-dimensional S-parameter container

        S-parameter data is expected to be of shape (..., nports, nports, nfreqs)


        Args:
            mdim_data (np.ndarray): _description_
            axes (list[np.ndarray]): _description_
        """        
        self._sdata: np.ndarray = mdim_data
        self._axes: list[np.ndarray] = axes
        self._shape: tuple[int] = self._sdata.shape
        self._extra_dims: int = len(self._shape) - 3
        self.nports: int = self._shape[-2]
        self.nfreqs: int = self._shape[-1]

    @property
    def _outer_axes(self) -> tuple:
        return tuple([ax for ax in self._axes[:self._extra_dims]])
    
    @property
    def _outer_slice(self) -> tuple:
        return tuple([slice(None) for _ in range(self._extra_dims)])
    
    def S(self, p1: int, p2: int) -> np.ndarray:
        """Get S-parameter S[p1, p2] counting from 1, S(1,1) is S11"""
        # check if p1 and p2 are valid ports
        if p1 < 1 or p1 > self.nports:
            raise ValueError(f"Port {p1} out of range")
        if p2 < 1 or p2 > self.nports:
            raise ValueError(f"Port {p2} out of range")
        return np.squeeze(self._sdata[self._outer_slice + (p1-1, p2-1, None)])
    
    def __call__(self, i1: int, i2: int) -> np.ndarray:
        return self.S(i1, i2)
    
    def intS(self, p1: int, p2: int) -> Interpolator:
        """Interpolates the S-parameter data for the given port combinations"

        Args:
            p1 (int): Port number
            p2 (int): Port number
        
        Returns:
            Interpolator: Interpolator object.

        Example:
        --------
        >>> S = NDSparameters(data, axes)
        >>> S.intS(1, 2)(f1, f2)
        """
        if any([ax.shape[0] > 1 for ax in self._outer_axes]):
            logger.warning('Parallel axis found in outer dimensions. Defaulting to first detected axis')
        axes = [np.squeeze(ax[0,:]) for ax in self._outer_axes] + [self._axes[-1],]
        return Interpolator(RegularGridInterpolator(axes, self.S(p1, p2), bounds_error=False, fill_value=None))
    



#            __   ___  __      __   ___       ___       __   __         ___      ___ 
# |  | |\ | |  \ |__  |__)    |  \ |__  \  / |__  |    /  \ |__)  |\/| |__  |\ |  |  
# \__/ | \| |__/ |___ |  \    |__/ |___  \/  |___ |___ \__/ |     |  | |___ | \|  |  
# -------------------------------------------

from numba import njit, prange, c16, i8, f8, typed, types, typeof, jit
import numba_progress as nbp
from numba_progress.progress import ProgressBarType
import numpy as np

#placeholder decorator that instead of compiling with numba prints the type using typeof of the called arguments and then executes the function
# def njit(*args, **kwargs):
#     def wrapper(func):
#         def wrapped(*args, **kwargs):
#             return func(*args, **kwargs)
#         return wrapped
#     return wrapper

# def jit(*args, **kwargs):
#     def wrapper(func):
#         def wrapped(*args, **kwargs):
#             return func(*args, **kwargs)
#         return wrapped
#     return wrapper

# Diode data structure
# Tuple[nodeid1, nodeid2, Idata, Vdata]

#### THESE ARE EXAMPLES OF THE DATA STRUCTURES USED IN THE FUNCTION AND ONLY INTENDED TO DEFINE NUMBA TYPE HINTS

_DIODE_EXAMPLE = [(1,2,np.array([1.,2.,3.]),np.array([1.,2.,3.])), (4,5,np.array([1.,2.,3.]),np.array([1.,2.,3.]))]
_LC_EXAMPLE = [(0,1,2,0,1e-12), (1,4,5,0,1e-12)]
_VSIGNALS_EXAMPLE = [(1,np.linspace(0,1,12),np.linspace(0,1,12)), (3,np.linspace(0,1,13),np.linspace(0,1,12))]

@njit(types.Tuple((f8[:,:],f8[:]))(f8, f8, f8[:], f8[:]), cache=True, fastmath=True)
def compute_diode(v1, v2, Idata, Vdata):
    EPS = 1e-15
    dV = v1-v2
    I0 = np.interp(dV, Vdata, Idata)
    i2 = np.interp(dV+EPS, Vdata, Idata)
    i1 = np.interp(dV-EPS, Vdata, Idata)
    G0 = (i2-i1)/(2*EPS)
    G = np.array([[1,-1],[-1,1]]) * G0
    I = np.array([-1,1]) * (I0 - G0*dV)
    return G, I

@njit(f8[:,:](f8[:,:], f8[:,:], i8[:]), cache=True, fastmath=True)
def add_submatrix(larger_matrix, smaller_matrix, indices):
    n = indices.shape[0]
    for i in range(n):
        for j in range(n):
            larger_matrix[indices[i], indices[j]] += smaller_matrix[i, j]
    return larger_matrix

@njit(types.Tuple((f8[:,:],f8[:]))(f8[:,:], f8[:], i8, i8, f8, f8, f8), cache=True, fastmath=True)
def implement_capacitor(A, x, i1, i2, C, Vold, dt):
    A[i1,i1] += C/dt
    A[i2,i2] += C/dt
    A[i1,i2] -= C/dt
    A[i2,i1] -= C/dt
    x[i1] += C*(Vold)/dt
    x[i2] -= C*(Vold)/dt
    return A, x

@njit(types.Tuple((f8[:,:],f8[:]))(f8[:,:], f8[:], i8, i8, i8, f8, f8, f8, f8), cache=True, fastmath=True)
def implement_inductor(A, x, i1, i2, vi, L, Vold, Iold, dt):
    A[i1, vi] += 1
    A[vi, i1] += 1
    A[i2, vi] += -1
    A[vi, i2] += -1
    A[vi, vi] += -L*2/dt
    x[vi] = -L*2*(Iold)/dt - Vold
    return A, x

@njit(f8[:,:](f8[:,:], f8[:], 
             i8, 
             i8, 
             typeof(_VSIGNALS_EXAMPLE), 
             f8[:], 
             i8, 
             typeof(_DIODE_EXAMPLE), 
             typeof(_LC_EXAMPLE),
             f8, 
             ProgressBarType), 
      cache=True, fastmath=True, nogil=True)
def solve_MNA_TRAN(Alin, Xlin, N, M, Vsignals, rampsteps, Niter, diodes, lcs,  alpha, pgb):
    """"""

    NM = N+M

    Vsignals = Vsignals[1:]

    time_series = Vsignals[0][2]
    dt = time_series[1] - time_series[0]

    #n_signal_sources = len(Vsignals)
    n_time_steps = Vsignals[0][2].shape[0]


    # Initialize the S-parameter matrix
    v_data_time = np.zeros((NM,n_time_steps))
    v_data = np.zeros((NM,))

    Amat = np.zeros_like(Alin)
    Xvec = np.zeros_like(Xlin)
    Xbase = np.zeros_like(Xlin)

    diodes = diodes[1:]
    
    extended_rampsteps = np.repeat(rampsteps, Niter)

    Xold = np.zeros_like(Xlin)
    Xnew = np.zeros_like(Xlin)

    for it in range(n_time_steps):
        Xold = 1*Xbase
        Xnew = 1*Xlin
        
        for source in Vsignals:
            Xnew[source[0]] = source[1][it]
        
        Xdiff = Xnew - Xold
        
        # Reset voltage vector
        v_new = np.zeros((NM,))
        
        for r in extended_rampsteps:
            #for i in range(Niter):
            Amat = Amat*0 + Alin
            Xvec = Xold + Xdiff*r

            for diode in diodes:
                i1, i2 = diode[0], diode[1]
                (G, I) = compute_diode(v_data[i1], v_data[i2], diode[2], diode[3])
                Amat = add_submatrix(Amat, G, np.array([i1,i2]))
                Xvec[i1] = I[0]
                Xvec[i2] = I[1]
            
            for lc in lcs[1:]:
                tp, i1, i2,vi,lcval = lc
                vold = v_data_time[i1,it-1] - v_data_time[i2,it-1]
                iold = v_data_time[vi,it-1]

                if tp == 0:
                    Amat, Xvec = implement_capacitor(Amat, Xvec, i1, i2, lcval, vold, dt)
                if tp == 1:
                    Amat, Xvec = implement_inductor(Amat, Xvec, i1, i2, vi, lcval, vold, iold, dt)
            
            Amat = Amat + np.eye(NM)*1e-15
            #Xvec = Xvec + np.random.randn(NM)*1e-15
            if np.all(Xvec[1:] == 0):
                v_new = v_data
            else:
                v_new[1:] = np.linalg.solve(Amat[1:,1:], Xvec[1:])
                #v_new[1:] = np.linalg.lstsq(Amat[1:,1:], Xvec[1:])[0]
                #v_new[1:] = np.linalg.pinv(Amat[1:,1:]) @ Xvec[1:]

            v_data = v_data + alpha*(v_new-v_data)

        v_data_time[:,it] = v_data
        if it % 2 == 0:
            pgb.update(2)
        #pgb.update(1)
    return v_data_time

@njit(f8[:,:](f8[:,:], f8[:], i8, i8, typeof(_VSIGNALS_EXAMPLE), f8[:], i8, typeof(_DIODE_EXAMPLE), f8), 
      cache=True, parallel=True, fastmath=True)
def solve_MNA_TRAN_nopgb(Alin, Xlin, N, M, Vsignals, rampsteps, Niter, diodes, alpha):
    """"""

    NM = N+M

    Vsignals = Vsignals[1:]
    #n_signal_sources = len(Vsignals)
    n_time_steps = Vsignals[0][2].shape[0]

    # Initialize the S-parameter matrix
    v_data_time = np.zeros((N,n_time_steps))
    v_data = np.zeros((N,))

    Amat = np.zeros_like(Alin)
    Xvec = np.zeros_like(Xlin)
    Xbase = np.zeros_like(Xlin)

    diodes = diodes[1:]

    for it in range(n_time_steps):
        Xbase = 0*Xbase + Xlin
        v_data = 0*v_data

        for source in Vsignals:
            i1 = source[0]
            Vdata = source[1]
            Xbase[i1] = Vdata[it]
            
        for r in rampsteps:
            for i in range(Niter):
                Amat = Amat*0 + Alin
                Xvec = Xvec*0 + Xbase*r

                for diode in diodes:
                    i1, i2 = diode[0], diode[1]
                    (G, I) = compute_diode(v_data[i1], v_data[i2], diode[2], diode[3])
                    Amat = add_submatrix(Amat, G, np.array([i1,i2]))
                    Xvec[i1] = I[0]
                    Xvec[i2] = I[1]

                # Reset voltage vector
                v_new = np.zeros((NM,))
                
                if np.all(Xvec[1:] == 0):
                    v_new = v_data
                else:
                    v_new[1:] = np.linalg.solve(Amat[1:,1:], Xvec[1:])

                
                # Check convergence
                maxchange = np.max(np.abs(v_new[:N] - v_data[:N]))
                maxV = np.max(np.abs(v_new[:N]))

                #if maxchange < eaps + erel*maxV:
                #    v_data = v_new[:N]
                #    break
                v_data = v_data + alpha*(v_new[:N] - v_data)
                #v_data = v_new[:N]
        v_data_time[:,it] = v_data
    return v_data_time


import numpy as np
from typing import List, Callable, Tuple, Literal
from ..node import Node, MatSource
from ..numeric import SimParam, enforce_simparam, enforce_value
from enum import Enum

_TWO_MAT = np.array([[1.0,-1.0],[-1.0,1.0]])
_TWO_MAT_C = np.array([[1.0,-1.0],[-1.0,1.0]]).astype(np.complex128)

TEN_POWERS = {
    -12: "p",
    -9: "n",
    -6: "u",
    -3: "m",
    0: "",
    3: "k",
    6: "M",
    9: "T",
    12: "P",
}

class SimulationType(Enum):
    AC = 1
    DC = 2
    TRANS = 3
    SP = 4

def _get_power(number: float):
    ''' Splits a float into a base value and the power of 10 in groups of 1000.
    
    >>> _get_power(4532)
    (4.532, 3)
    '''
    N_thousands = np.floor(np.log10(np.abs(number)) / 3) * 3
    N_thousands = min(12, max(-12, N_thousands))
    base_number = number / (10**N_thousands)
    return base_number, N_thousands

def _format_value_units(value: float, unit: str) -> str:
    """ Formats a value with units for display. 
    
    >>> format_value_units(4532, 'Hz')
    '4.53 kHz'
    """
    
    v, p = _get_power(enforce_value(value))
    return f"{v:.2f} {TEN_POWERS[p]}{unit}"

class BaseComponent:

    supported_simulations: List[SimulationType] = []
    
    PRINT_FREQ: float = 1e0

    AC_sources: int = 0
    DC_sources: int = 0
    TRANS_sources: int = 0
    SP_sources: int = 0

    AC_INTNODES: int = 0
    DC_INTNODES: int = 0
    TRANS_INTNODES: int = 0
    SP_INTNODES: int = 0

    _is_linear: bool = True

    def __init__(self):
        self.nodes: List[Node] = []
        self.nterminals: int = 0
        self.sources: List[MatSource] = []
        self.internal_nodes: List[Node] = []

        self.component_name: str = 'BaseComponent'
        self.unit: str = ''
        self.value: float = 'Undefined'

        self.metadata: dict[str,str] = {}

    def __repr__(self):
        return f'{self.component_name}[{_format_value_units(self.value, self.unit)}]'

    def __str__(self):
        return self.__repr__()
    
    def summary(self) -> str:
        ''' Returns a multiline string summarizing component data. '''
        base = dict(name=self.component_name, value=self.value, unit=self.unit)
        base.update(self.metadata)

        lines = [f'{base["name"]} ({_format_value_units(base["value"], base["unit"])})']
        del base["name"]
        del base["value"]
        del base["unit"]
        if base:
            lines.append('-'*len(lines[0]))
            for key, value in base.items():
                lines.append(f'    {key}: {value}')
        return '\n'.join(lines)
    
    def set_metadata(self, **kwargs):
        self.metadata.update(kwargs)
        
    @property
    def all_nodes(self) -> List[Node]:
        return self.nodes + self.internal_nodes
    
    def is_linear(self, simulation_type: SimulationType) -> bool:
        return self._is_linear
    
    def n_requested_sources(self, simulation_type: SimulationType) -> int:
        if simulation_type == SimulationType.AC:
            return self.AC_sources
        if simulation_type == SimulationType.DC:
            return self.DC_sources
        if simulation_type == SimulationType.TRANS:
            return self.TRANS_sources
        if simulation_type == SimulationType.SP:
            return self.SP_sources
    
    def n_requested_internal_nodes(self, simulation_type: SimulationType) -> int:
        if simulation_type == SimulationType.AC:
            return self.AC_INTNODES
        if simulation_type == SimulationType.DC:
            return self.DC_INTNODES
        if simulation_type == SimulationType.TRANS:
            return self.TRANS_INTNODES
        if simulation_type == SimulationType.SP:
            return self.SP_INTNODES
    
    def indices(self) -> List[int]:
        return [node.index for node in self.nodes] + [source.index for source in self.sources]
    
    def set_sources(self, sources: List[MatSource]):
        self.sources = sources

    def set_internal_nodes(self, nodes: List[Node]):
        self.internal_nodes = nodes

    @property
    def mat_slice(self) -> Tuple[np.ndarray, np.ndarray]:
        ids = np.array(self.indices())
        return np.ix_(ids,ids)
    
    @property
    def vec_slice(self) -> np.ndarray:
        return np.ix_(self.indices())

    def AC_matrix_compiler(self) -> Callable:
        return None
    
    def AC_source_compiler(self) -> Callable:
        return None
    
    def DC_matrix_compiler(self) -> Callable:
        return None
    
    def DC_source_compiler(self) -> Callable:
        return None
    
    def TRANS_matrix_compiler(self) -> Callable:
        return None
    
    def TRANS_source_compiler(self) -> Callable:
        return None
    
    def SP_matrix_compiler(self) -> Callable:
        return None
    
    def SP_source_compiler(self) -> Callable:
        return None
    



########################################################################################
##
##    Numba accelerated S-parameter calculation for RF networks
##    This code is part of the Heavi framework
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


from numba import njit, prange, c16, i8, f8
import numba_progress as nbp
from numba_progress.progress import ProgressBarType
import numpy as np


#  __   __             ___  __   __  
# /__` /  \ |    \  / |__  |__) /__` 
# .__/ \__/ |___  \/  |___ |  \ .__/ 
# -------------------------------------------


@njit(cache=True, parallel=True, fastmath=True)
def solve_MNA_RF(As, Zs, port_indices, frequencies, nnodes, nsources, progress_object):
    """
    Compute the S-parameter matrix for an RF network using Numba for acceleration.

    Parameters
    ----------
    Is : numpy.ndarray
        Current sources, complex-valued array of shape (n_nodes, n_ports, n_freqs).
    Ys : numpy.ndarray
        Admittance matrices, complex-valued array of shape (n_nodes, n_nodes, n_freqs).
    Zs : numpy.ndarray
        Source impedances, complex-valued array of shape (n_nodes, n_freqs).
    port_indices : numpy.ndarray
        Indices of the nodes corresponding to the ports of interest, integer array of shape (n_ports,).
    frequencies : numpy.ndarray
        Frequencies, float-valued array of shape (n_freqs,).

    Returns
    -------
    S_parameters : numpy.ndarray
        S-parameter matrix, complex-valued array of shape (n_ports, n_ports, n_freqs).
    """
    M = nsources
    N = nnodes
    nports = port_indices.shape[0]
    p_index = port_indices[:,0]
    p_out_nodes = port_indices[:,1]
    p_int_nodes = port_indices[:,2]
    p_gnd_nodes = port_indices[:,3]

    num_freqs = len(frequencies)
    NM = N+M
    
    # Initialize the S-parameter matrix
    v_data = np.zeros((N, num_freqs), dtype=np.complex128)
    S_parameters = np.zeros((nports, nports, num_freqs), dtype=np.complex128)
    # x vector placeholder

    for i in range(nports):

        active_port_index = p_index[i]

        for freq_idx in prange(num_freqs):
            # Reset voltage vector
            x = np.zeros((NM,), dtype=np.complex128)
            z = np.zeros((NM,), dtype=np.complex128)
            # Set 1V at the active port node
            z[N + active_port_index] = 1
            # Solve the system of equations for Vh[1:]
            x[1:] = np.linalg.lstsq(As[1:,1:,freq_idx], z[1:])[0].astype(np.complex128)
            #x[1:] = np.linalg.solve(As[1:,1:,freq_idx], z[1:])
            #x[1:] = np.linalg.pinv(subA[:,:,freq_idx]) @ z
            Z_in = Zs[active_port_index, freq_idx]

            v_data[:,freq_idx] = x[:N]

            Vi1 = x[p_out_nodes[i]]
            Vi2 = x[p_gnd_nodes[i]]
            Vi3 = x[p_int_nodes[i]]

            for j in range(nports):
                Z_out = Zs[p_index[j], freq_idx]
                # Compute numerator and denominator for S-parameter calculation
                Vo1 = x[p_out_nodes[j]]
                Vo2 = x[p_gnd_nodes[j]]
                Vo3 = x[p_int_nodes[j]]
                numerator = (Vo1 - Vo2) + np.conj(Z_out) * (Vo1 - Vo3) / Z_out
                denominator = (Vi1 - Vi2) -  Z_in * (Vi1-Vi3) / Z_in
                # Compute S-parameter
                S_parameters[j, i, freq_idx] = (numerator / denominator) * np.sqrt(np.abs(np.real(Z_in))) / np.sqrt(np.abs(np.real(Z_out)))
            
            progress_object.update(1)

    return v_data, S_parameters


@njit(cache=True, parallel=True, fastmath=True)
def solve_MNA_RF_nopgb(As, Zs, port_indices, frequencies, nnodes, nsources):
    """
    Compute the S-parameter matrix for an RF network using Numba for acceleration.

    Parameters
    ----------
    Is : numpy.ndarray
        Current sources, complex-valued array of shape (n_nodes, n_ports, n_freqs).
    Ys : numpy.ndarray
        Admittance matrices, complex-valued array of shape (n_nodes, n_nodes, n_freqs).
    Zs : numpy.ndarray
        Source impedances, complex-valued array of shape (n_nodes, n_freqs).
    port_indices : numpy.ndarray
        Indices of the nodes corresponding to the ports of interest, integer array of shape (n_ports,).
    frequencies : numpy.ndarray
        Frequencies, float-valued array of shape (n_freqs,).

    Returns
    -------
    S_parameters : numpy.ndarray
        S-parameter matrix, complex-valued array of shape (n_ports, n_ports, n_freqs).
    """
    M = nsources
    N = nnodes
    nports = port_indices.shape[0]
    p_index = port_indices[:,0]
    p_out_nodes = port_indices[:,1]
    p_int_nodes = port_indices[:,2]
    p_gnd_nodes = port_indices[:,3]

    num_freqs = len(frequencies)
    NM = N+M
    
    # Initialize the S-parameter matrix
    v_data = np.zeros((N, num_freqs), dtype=np.complex128)
    S_parameters = np.zeros((nports, nports, num_freqs), dtype=np.complex128)
    # x vector placeholder

    for i in range(nports):

        active_port_index = p_index[i]

        for freq_idx in prange(num_freqs):
            # Reset voltage vector
            x = np.zeros((NM,), dtype=np.complex128)
            z = np.zeros((NM,), dtype=np.complex128)
            # Set 1V at the active port node
            z[N + active_port_index] = 1
            # Solve the system of equations for Vh[1:]
            x[1:] = np.linalg.lstsq(As[1:,1:,freq_idx], z[1:])[0].astype(np.complex128)
            #x[1:] = np.linalg.solve(As[1:,1:,freq_idx], z[1:])
            #x[1:] = np.linalg.pinv(subA[:,:,freq_idx]) @ z
            Z_in = Zs[active_port_index, freq_idx]

            v_data[:,freq_idx] = x[:N]

            Vi1 = x[p_out_nodes[i]]
            Vi2 = x[p_gnd_nodes[i]]
            Vi3 = x[p_int_nodes[i]]

            for j in range(nports):
                Z_out = Zs[p_index[j], freq_idx]
                # Compute numerator and denominator for S-parameter calculation
                Vo1 = x[p_out_nodes[j]]
                Vo2 = x[p_gnd_nodes[j]]
                Vo3 = x[p_int_nodes[j]]
                numerator = (Vo1 - Vo2) + np.conj(Z_out) * (Vo1 - Vo3) / Z_out
                denominator = (Vi1 - Vi2) -  Z_in * (Vi1-Vi3) / Z_in
                # Compute S-parameter
                S_parameters[j, i, freq_idx] = (numerator / denominator) * np.sqrt(np.abs(np.real(Z_in))) / np.sqrt(np.abs(np.real(Z_out)))
            

    return v_data, S_parameters


########################################################################################
##
##    Resistor component Definition
##    This file contains the definition of the Resistor component.
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################


from .component import BaseComponent, Node, SimParam, enforce_simparam, SimulationType, _TWO_MAT
from typing import List, Callable
import numpy as np


#  __             __   __  
# /  ` |     /\  /__` /__` 
# \__, |___ /~~\ .__/ .__/ 
# -------------------------------------------


class Resistor(BaseComponent):
    '''
    """The Resistor component models a simple resistor component without paracitic properties.

    Returns:
        Resistor: Resistor component
    """    '''
    supported_simulations: List[SimulationType] = [SimulationType.AC,
                                                   SimulationType.DC, 
                                                   SimulationType.SP,
                                                   SimulationType.TRANS]
    
    def __init__(self, node1: Node, node2: Node, resistance: float | SimParam):
        super().__init__()
        self.nodes = [node1, node2]
        self.nterminals = 2
        self.R: SimParam = enforce_simparam(resistance)
        self.G: SimParam = enforce_simparam(lambda f: 1/self.R(f))
        
        self.component_name = 'Resistor'
        self.unit = 'Ohm'
        self.value = resistance
    
    @property
    def mat_slice(self) -> tuple[np.ndarray, np.ndarray]:
        ids = np.array([self.nodes[0].index, self.nodes[1].index])
        return np.ix_(ids, ids)
    
    def generic_matrix_compiler(self) -> Callable:
        slc = self.mat_slice

        def compiler(matrix: np.ndarray, f: float) -> np.ndarray:
            matrix[slc] += self.G(f) * _TWO_MAT[:,:,np.newaxis]
            return matrix
        
        return compiler
    
    def SP_matrix_compiler(self):
        return self.generic_matrix_compiler()
    
    def DC_matrix_compiler(self):
        slc = self.mat_slice

        def compiler(matrix: np.ndarray) -> np.ndarray:
            matrix[slc] += self.G(0.0) * _TWO_MAT
            return matrix
        
        return compiler
    
    def AC_matrix_compiler(self):
        return self.generic_matrix_compiler()
    
    def TRANS_matrix_compiler(self):
        slc = self.mat_slice

        def compiler(matrix: np.ndarray) -> np.ndarray:
            matrix[slc] += self.G(0.0) * _TWO_MAT
            return matrix
        
        return compiler
    


from .component import BaseComponent, Node, SimParam, enforce_simparam, SimulationType, _TWO_MAT_C, _TWO_MAT
from typing import List, Callable
import numpy as np
from .numeric import stack, cstack

class Capacitor(BaseComponent):
    supported_simulations: List[SimulationType] = [SimulationType.AC,
                                                   SimulationType.DC, 
                                                   SimulationType.SP,
                                                   SimulationType.TRANS]
    
    def __init__(self, node1: Node, node2: Node, capacitance: float | SimParam):
        super().__init__()
        self.nodes = [node1, node2]
        self.nterminals = 2
        self.C: SimParam = enforce_simparam(capacitance)
        
        self.component_name = 'Capacitor'
        self.unit = 'F'
        self.value = capacitance
    
    @property
    def mat_slice(self) -> tuple[np.ndarray, np.ndarray]:
        ids = np.array([self.nodes[0].index, self.nodes[1].index])
        return np.ix_(ids, ids)
    
    def SP_matrix_compiler(self):
        row,col = self.mat_slice

        def compiler(matrix: np.ndarray, f: float) -> np.ndarray:
            matrix[row,col,:] += cstack(_TWO_MAT_C, 1j*2*np.pi*f*self.C(f) ,0*_TWO_MAT_C)
            return matrix
        
        return compiler
    
    def DC_matrix_compiler(self):
        slc = self.mat_slice

        def compiler(matrix: np.ndarray) -> np.ndarray:
            matrix[slc] += 0 * _TWO_MAT
            return matrix
        
        return compiler
    
    def AC_matrix_compiler(self):
        return self.SP_matrix_compiler()
    